;;;  Author:  Andy Philips (abp@goedel.arc.nasa.gov) 

#|

Circular lists are extremely useful when each member of the list is
``aware'' of its presence in that list.  Members of circular lists
defined in this manner, may be added and deleted in constant time.  Of
course, by reducing the computation time associated with membership
tests, extra memory is needed to maintain the lists.

|#


(defun construct-circular-list (op preds nec)
  (setf (operator-pre op)
	(mapcar #'(lambda (x) (make-pred :op op :goal x)) preds))
  (dolist (item (operator-pre op))
	  (circular-add nec item)))


(defun circular-add (head new)
  (declare (optimize (safety 0) (speed 3))
	   (type pred new))
  (circular-delete new)
  (let ((member (car head)))
    (declare (type pred member))
    (if member
	(setf (pred-bef new) member
	      (pred-aft new) (pred-aft member)
	      (pred-aft member) new
	      (pred-bef (pred-aft new)) new)
      (setf (pred-bef new) new
	    (pred-aft new) new))
    (setf (pred-list new) head)
    (rplaca head new)))


(defun circular-delete (member)
  (declare (optimize (safety 0) (speed 3))
	   (type pred member))
  (if (pred-list member)
      (let ((temp (pred-bef member)))
	(declare (type pred temp))
	(cond ((not (eq temp member))
	       (setf (pred-bef (pred-aft member)) (pred-bef member)
		     (pred-aft (pred-bef member)) (pred-aft member)
		     (pred-aft member) member
		     (pred-bef member) member)
	       (rplaca (pred-list temp) temp)
	       (setf (pred-list member) nil))
	      (t
	       (rplaca (pred-list member) nil)
	       (setf (pred-list member) nil))))))


#|
;; Warning these may not be members the same list or dire straits will
;; result.

(defun circular-splice (cl1 cl2)
  (cond ((null cl1) cl2)
	((null cl2) cl1)
	((eq cl2 cl1) cl2)
	(t (setf (pred-bef (pred-aft cl1)) (pred-bef cl2)
		 (pred-aft (pred-bef cl2)) (pred-aft cl1)
		 (pred-bef cl2) cl1
		 (pred-aft cl1) cl2)
	   cl1)))


(defun circular-search-and-delete (item cl)
  (let (garbage)
    (setf garbage nil)
    (loop with temp = nil
	  while (and cl (equal (pred-goal cl) item))
	  do (setf cl (circular-delete (setf temp cl))
		   garbage (circular-add garbage temp)))
    (if cl
	(loop with i = cl
	      with temp = nil
	      if (equal (pred-goal i) item)
	      do (setf i (circular-delete (setf temp i))
		       garbage (circular-add garbage temp))
	      until (or (null i) (eq cl (setf i (pred-aft i))))))
    (values cl garbage)))

|#