;;;  Contains: Structure definitions for the planners
;;;  Author:  Andy Philips 

(proclaim '(optimize (safety 0) (speed 3)))

(defvar *concise-class* t)

#| OPERATOR-CLASS

This contains a description of the operators that are availble for
planning.

|#

(defstruct (operator-class (:print-function
			    (lambda (struct stream depth)
			      (print-operator-class struct stream depth))))
  (name nil)
  (static nil)	;; Conjunction of static predicates that enable the operator
  (pre nil)	;; Conjunction of predicates that enable the operator
  (del nil)	;; Conjunction of predicates that operator deletes
  (add nil))	;; Conjunction of predicates that operator adds


(defstruct (pred (:print-function
		  (lambda (struct stream depth)
		    (print-pred struct stream depth))))
  (goal nil)	;; The goal this predicate describes
  (? 'maybe)	;; If the goal is satisfied by an add somewhere (T = satisfied)
  (op nil)	;; The operator this goal is linked with
  (list nil)	;; The circular list this pred is a member of
  (bef nil)	;; Circular List link (before)
  (aft nil))	;; Circular List link (after)
  

#| OPERATOR

This is a particular instance of the class of an operator.  This
instance is necessary so that a plan may contain more than one
instance of the same operator and have that operator correctly ordered
with repect to the other operators within the plan.

|#

(defstruct (operator (:print-function
		      (lambda (struct stream depth)
		       (print-operator struct stream depth))))
  (class nil)		;; The class of this operator
  (instance nil)	;; The name of this operator
  (pre nil)		;; List of predicates in the Left Hand Side
  (preds nil)		;; Those instances that preceed operator
  (succs nil)		;; Those that succeed this operator
  (mark  0)		;; To mark partial orders, number of visits
  (bef nil)		;; Doubly-linked linearized UA plan: the previous step
  (aft nil)		;; Doubly-linked linearized UA plan: the following step
  (graphics nil)	;; Slot for graphics representation
  (disam nil))		;; Used to stack the disambiguation process

#|
operators will look like this:

<class
instance 4
preds  '(3)
succs    nil>

<class
instance 3
preds  '(2)
succs   '(4)>




Insert 5 between 3 & 4:

<class
instance 3
preds '(2)
succs  '(5 4)>

<class
instance 4
preds '(5 3)
succs  nil>

|#

(defun print-operator-class (struct stream depth)
  (unless depth (setf depth 1))
  (if (or *concise-class* (> depth 1))
      (princ (operator-class-name struct) stream)
    (format stream "#s(OPERATOR-CLASS :NAME ~a :STATIC ~a :PRE ~a :DEL ~a :ADD ~a)"
	    (operator-class-name struct)
	    (operator-class-static struct)
	    (operator-class-pre struct)
	    (operator-class-del struct)
	    (operator-class-add struct))))


(defun print-operator (struct stream depth)
  (unless depth (setf depth 1))
  (if (< depth 2)
      (progn
	(format stream "#<~a " (operator-instance struct))
	(let ((op-class (operator-class struct)))
	  (format stream "~a" (if op-class (operator-class-name op-class))))
	(print-list-ops (operator-preds struct) " Preds: " stream)
	(print-list-ops (operator-succs struct) " Succs: " stream)
	(format stream ">"))
    (princ (operator-instance struct) stream)))

#|
   This displays the operator marks and the bef/aft pointers.

(defun print-operator (struct stream depth)
  (unless depth (setf depth 1))
  (cond ((not (< depth 2))
	 (princ (operator-instance struct) stream))
	(*debug*
	 (format stream "#<~a" (operator-instance struct))
	 (if (operator-bef struct)
	     (format stream " bef: ~a"
		     (operator-instance (operator-bef struct))))
	 (if (operator-aft struct)
	     (format stream " aft: ~a"
		     (operator-instance (operator-aft struct))))
	 (format stream ">"))
	(t
	 (format stream "#<~a " (operator-instance struct))
	 (let ((op-class (operator-class struct)))
	   (format stream "~a" (if op-class (operator-class-name op-class))))
	 (print-list-ops (operator-preds struct) " Preds: " stream)
	 (print-list-ops (operator-succs struct) " Succs: " stream)
	 (format stream ">"))))
|#



(defun print-list-ops (l string s)
  (format s "~a" string)
  (cond (l
	 (format s "(~a" (operator-instance (car l)))
	 (dolist (op (cdr l)) (format s " ~a" (operator-instance op)))
	 (format s ")"))
	(t (format s "NIL"))))


(defun print-pred (struct stream depth)
  (unless depth (setf depth 1))
  (cond ((< depth 2)
	 (princ "#cl(" stream)
	 (loop with i = struct
	   do (print-pred-sub i stream)
	      (setf i (pred-aft i))
	   if (null i) do (princ "-->NIL" stream)
	   until (or (eq struct i) (null i))
	   do (princ " " stream))
	 (princ ")" stream))
	(t (print-pred-sub struct stream))))


(defun print-pred-sub (struct stream)
  (format stream "~a:" (pred-goal struct))
  (if (operator-p (pred-op struct))
      (format stream "~a" (operator-instance (pred-op struct)))
      (princ "NO-OP" stream)))
