;;;  Contains: Graphs runs of the problems for comparison of data
;;;  Author:  Andy Philips (abp@goedel.arc.nasa.gov)
;;;  Created: Mon Oct 28 10:44:42 1991


#| How to use this file:

(setf *data* (full-test '(<list of the problems you would like to run>) <number_of_iterations>))

(create-function '(chunk-depth chunk) '(log (chunk-nodes chunk) 9)
		 "Log Nodes vs. Depth" "Depth" "Log Nodes")

(setf *set* (make-a-set *data* (find <fn_number> *functions* :key #'CHUNK-FUNCTION-NUMBER)))

(view-them *set* '(ua to) :means t :data nil :lines t)

|#


(setf *graphics-on* nil)
(setf *out-stream* nil)
(defvar *brief-chunks* t)
(gensym "Lisp")
(gensym 1000)
(defvar *functions* nil)
(defvar *num-functions* 0)


(defstruct (chunk (:print-function
		   (lambda (struct stream depth)
		     (print-chunk struct stream depth))))
  (nodes 0)	;; Number of nodes visited
  (time 0)	;; CPU time measured in seconds
  (branch 0)	;; Operator branching factor for this chunk
  (depth 0)	;; Depth level for this problem
  (problem 0))	;; The particular problem


(defstruct (chunk-planner (:print-function
			   (lambda (struct stream depth)
			     (print-chunk-planner struct stream depth))))
  (type nil)	;; The type of planner used (UA UAP TO TOP)
  (chunks nil))	;; Chunks produced by that planner


(defstruct chunk-function
  (fnx #'(lambda (chunk) 0))	;; Finds the X value for point
  (fny #'(lambda (chunk) 0))	;; Finds the Y value for point
  (name "No name")		;; Name for this function
  (namex "X")			;; Name for X axis
  (namey "Y")			;; Name for Y axis
  (number 0)			;; Function number
  (fn #'(lambda (c) '(0 . 0))))	;; The function that produces a point from a chunk


(defstruct stat
  (item nil)		;; Particular class of number
  (total 0)		;; The total for this class
  (count 0)		;; Number of entries in class
  (mean 0)		;; Mean for the Class
  (deviation 0))	;; Standard deviation for the Class


(defstruct sub-set
  (points nil)		;; Points in this data sub-set
  (max-x (caar points))	;; The Maximum X value in the sub-set
  (min-x (caar points))	;; The Minimum X value in the sub-set
  (max-y (cdar points))	;; The Maximum Y value in the sub-set
  (min-y (cdar points))	;; The Minumum Y value in the sub-set
  (stats nil)		;; Statistics for this sub-set
  (cp nil)		;; Data source for these points
  (cf nil))		;; Chunk function which created these points


(defun print-chunk (struct stream depth)
  (unless depth (setf depth 1))
  (cond ((= depth 1)
	 (format stream "#s(CHUNK :NODES ~a :TIME ~a :BRANCH ~
                         ~a :DEPTH ~a :PROBLEM ~a)"
		 (chunk-nodes struct) (chunk-time struct) (chunk-branch struct)
		 (chunk-depth struct) (chunk-problem struct)))
	(*brief-chunks*
	 (format stream "<~a ~a ~a ~a>"
		 (chunk-nodes struct)  (chunk-time struct)
		 (chunk-branch struct) (chunk-depth struct)))
	(t
	 (format stream "<N:~a T:~a B:~a D:~a P:~a>"
		 (chunk-nodes struct) (chunk-time struct) (chunk-branch struct)
		 (chunk-depth struct) (chunk-problem struct)))))


(defun print-chunk-planner (struct stream depth)
  (format stream "#s(~a ~a)" (chunk-planner-type struct)
	  (chunk-planner-chunks struct)))


(defun check-test-a-problem (pr)
  (let ((p '(uaf tof cuaf mtof)))
    (loop for fn in p
	  collect (cons pr (cons fn (funcall fn))))))


(defun check-problem-set (list-of-problems)
  (let ((data1 (loop with temp = nil
		     for pr in list-of-problems
		     do (problem pr)
		        (setf temp (check-test-a-problem pr))
			(print temp) (terpri)
		     nconc temp))
	(data2 (loop with temp = nil
		     for pr in list-of-problems
		     do (problem pr)
		        (decf *depth-level*)
			(format t "Setting Depth Level to ~A~%" *depth-level*)
		        (setf temp (check-test-a-problem pr))
			(print temp) (terpri)
		     nconc temp))
	(ok? t))
    (loop for (pr fn truth nodes time) in data1
	  if (not truth)
	  do (progn (format t "Depth Too Small: Problem ~A, Planner ~A, Nodes ~A~%"
		     pr fn nodes)
		    (setf ok? nil))
	  if (zerop nodes)
	  do (progn (format t "No Nodes Expand: Problem ~A, Planner ~A, Nodes ~A~%"
		     pr fn nodes)
		    (setf ok? nil)))
    (loop for (pr fn truth nodes time) in data2
	  if truth
	  do (progn (format t "Depth Too Large: Problem ~A, Planner ~A, Nodes ~A~%"
		     pr fn nodes)
		    (setf ok? nil))
	  if (zerop nodes)
	  do (progn (format t "No Nodes Expand: Problem ~A, Planner ~A, Nodes ~A~%"
		     pr fn nodes)
		    (setf ok? nil)))
    (if ok? (format t "~%...done testing.  Everything OK!~%")
      (format t "~%...done testing.  Got Problems!~%"))))


(defun output-to-file (list-of-chunk-planners file-name)
  (with-open-file
   (outfile file-name :direction :output)
   (format outfile "(")
   (loop for cp in list-of-chunk-planners
	 do (format outfile "(~A~%" (chunk-planner-type cp))
	    (loop for chunk in (chunk-planner-chunks cp)
		  do (format outfile " (~a ~a ~a ~a ~a)"
		     (chunk-nodes chunk) (chunk-time chunk) (chunk-branch chunk)
		     (chunk-depth chunk) (chunk-problem chunk)))
	    (format outfile ")"))
   (format outfile ")")))



(defun input-from-file (file-name)
  (let ((transfer (with-open-file
		   (infile file-name :direction :input) (read infile)))
	(holding nil))
    (setf holding
	  (mapcar #'(lambda (x)
		      (make-chunk-planner
		       :type (car x) :chunks
		       (mapcar #'(lambda (y)
				   (make-chunk
				    :nodes (nth 0 y) :time (nth 1 y)
				    :branch (nth 2 y) :depth (nth 3 y)
				    :problem (nth 4 y))) (cdr x))))
			  transfer))))



(defun doit (fn it pr)
  (loop with temp = nil
	for i from 1 to it
	do (setf temp (funcall fn))
	collect (make-chunk :nodes (nth 1 temp)
			    :time (nth 2 temp)
			    :branch *operator-branching*
			    :depth *depth-level*
			    :problem pr)))


(defun test-a-problem (it pr)
  (let ((p '((uaf ua) (tof to) (cuaf uap) (mtof top))))
    (loop for (fn nm) in p
	  collect (cons nm (doit fn it pr)))))


(defun full-test (list-of-problems it)
  (let ((data (loop with temp = nil
		    for pr in list-of-problems
		    do (problem pr)
		       (setf temp (test-a-problem it pr))
		       (print temp) (terpri)
		    nconc temp))
	(raw-data (list (make-chunk-planner :chunks nil :type 'ua)
			(make-chunk-planner :chunks nil :type 'uap)
			(make-chunk-planner :chunks nil :type 'to)
			(make-chunk-planner :chunks nil :type 'top))))
    (loop with cp = nil
	  for (type . rest) in data
	  do (setf cp (find type raw-data :key #'chunk-planner-type))
	     (setf (chunk-planner-chunks cp)
		   (nconc rest (chunk-planner-chunks cp)))
	  finally (return raw-data))))


#| How to use CREATE-FUNCTION:

(create-function '(chunk-depth chunk)		;; The X axis value (derived from CHUNK)
		 '(log (chunk-nodes chunk) 9)	;; The Y axis value (derived from CHUNK)
		 "Log Nodes vs. Depth"		;; What to call the graph
		 "Depth"			;; The name for the X axis
		 "Log Nodes")			;; The name for the Y axis

The newly created function is pushed onto the stack of functions.
Also, a function is compiled that when called on a chunk will produced
a dotted pair representing a point on the graph.

|#

(defun create-function (fnx fny name namex namey)
  (let ((cf (make-chunk-function :fnx fnx :fny :fny :name name :namex namex
				 :namey namey :number (incf *num-functions*))))
    (setf (chunk-function-fn cf) (compile nil `(lambda (chunk) (cons ,fnx ,fny))))
    (push cf *functions*)))


(defun list-functions (&optional (stream t))
  (loop for fn in *functions*
	do (format stream "~2d: <>~%" (chunk-function-number fn))))



(defun make-a-set (list-of-cp cf);; List of Chunk planners and Cunk function to apply
  (loop with sub = nil
	for cp in list-of-cp
	collect (setf sub (make-sub-set :points (mapcar (chunk-function-fn cf)
							(chunk-planner-chunks cp))
					:cf cf :cp cp))
	do (loop with stat = nil
		 for (x . y) in (sub-set-points sub)
		 if (< x (sub-set-min-x sub)) do (setf (sub-set-min-x sub) x)
		 if (> x (sub-set-max-x sub)) do (setf (sub-set-max-x sub) x)
		 if (< y (sub-set-min-y sub)) do (setf (sub-set-min-y sub) y)
		 if (> y (sub-set-max-y sub)) do (setf (sub-set-max-y sub) y)
		 do (setf stat (find x (sub-set-stats sub) :key #'stat-item))
		 if (null stat)
		 do (push (setf stat (make-stat :item x)) (sub-set-stats sub))
		 do (incf (stat-count stat)) (incf (stat-total stat) y))
	   (setf (sub-set-stats sub) (sort (sub-set-stats sub) #'< :key #'stat-item))
	   (loop for stat in (sub-set-stats sub)
		 do (setf (stat-mean stat) (/ (stat-total stat) (stat-count stat) 1.0)))
	   (loop with stat = nil
		 for (x . y) in (sub-set-points sub)
		 do (setf stat (find x (sub-set-stats sub) :key #'stat-item))
		    (incf (stat-deviation stat) (expt (- y (stat-mean stat)) 2)))
	   (loop for stat in (sub-set-stats sub)
		 do (setf (stat-deviation stat)
			  (sqrt (/ (stat-deviation stat) (expt (1- (stat-count stat)) 2)))))))


(defun output-a-plot (sub-title points stream)
  (format stream "~%~c~a~c~%" #\" sub-title #\")
  (loop for (x . y) in points do (format stream "draw ~a ~a~%" x y)))


(defun output-a-mean (sub-title stats stream)
  (format stream "~%~c~a Mean~c~%" #\" sub-title #\")
  (loop for stat in stats
	do (format stream "draw ~a ~a~%"
		   (stat-item stat) (stat-mean stat))))


(defun output-a-devi (sub-title stats stream)
  (format stream "~%~c~a Deviation~c~%" #\" sub-title #\")
  (loop for stat in stats
	do (format stream "draw ~a ~a~%draw ~a ~a~%" (stat-item stat)
		   (- (stat-mean stat) (stat-deviation stat)) (stat-item stat)
		   (+ (stat-mean stat) (stat-deviation stat)))))


(defun output-a-set (set list-of-planners stream lx ly data means devia lines file)
  (let ((cf (sub-set-cf (car set))) (bounds '(0 0 0 0)))
    (format stream "TitleText: ~a~%" (chunk-function-name cf))
    (format stream "XUnitText: ~A~%" (chunk-function-namex cf))
    (format stream "YUnitText: ~A~%" (chunk-function-namey cf))
    (if (not file)
	(format stream "Device: Postscript~%Disposition: To Device~%FileOrDev: alw~%")
        (format stream "Device: Postscript~%Disposition: To File~%FileOrDev: ~a~%" file))
    (loop initially (loop for i from 0 to 3 do (setf (nth i bounds) nil))
	  for sub in set
	  if (or (not (nth 0 bounds)) (< (nth 0 bounds) (sub-set-min-x sub)))
	  do (setf (nth 0 bounds) (sub-set-min-x sub))
	  if (or (not (nth 1 bounds)) (> (nth 1 bounds) (sub-set-max-x sub)))
	  do (setf (nth 1 bounds) (sub-set-max-x sub))
	  if (or (not (nth 2 bounds)) (< (nth 2 bounds) (sub-set-min-y sub)))
	  do (setf (nth 2 bounds) (sub-set-min-y sub))
	  if (or (not (nth 3 bounds)) (> (nth 3 bounds) (sub-set-max-y sub)))
	  do (setf (nth 3 bounds) (sub-set-max-y sub)))
    (cond (lx
	   (format stream "LogX: True~%")
	   (format stream "XLowLimit:  ~A~%" (log (nth 0 bounds) 10))
	   (format stream "XHighLimit: ~A~%" (log (nth 1 bounds) 10)))
	  (t 
	   (format stream "XLowLimit:  ~A~%" (nth 0 bounds))
	   (format stream "XHighLimit: ~A~%" (nth 1 bounds))))
    (cond (ly
	   (format stream "LogY: True~%")
	   (format stream "YLowLimit:  ~A~%" (log (nth 2 bounds) 10))
	   (format stream "YHighLimit: ~A~%" (log (nth 3 bounds) 10)))
	  (t 
	   (format stream "YLowLimit:  ~A~%" (nth 2 bounds))
	   (format stream "YHighLimit: ~A~%" (nth 3 bounds))))
    (format stream "Markers:   True~%")
    (if (or data (not lines)) (format stream "NoLines:   True~%"))
    (loop with sub = nil
	  for planner in list-of-planners
	  do (if (setf sub (find planner set :key
				 #'(lambda (x) (chunk-planner-type (sub-set-cp x)))))
		 (progn
		   (if data (output-a-plot planner (sub-set-points sub) stream))
		   (if means (output-a-mean planner (sub-set-stats sub) stream))
		   (if devia (output-a-devi planner (sub-set-stats sub) stream)))))))


(defun output-a-fn (fn title stats stream)
  (format stream "~%~c~a~c~%" #\" title #\")
  (loop for stat in stats
	do (format stream "draw ~a ~a~%"
		   (stat-item stat) (funcall fn (stat-item stat)))))


#|

Example of how to use:

(view-them *set-one* '(ua to) :means t :data nil :lines t :fns *means-two* :file t)

|#


(defun view-them (set list-of-planners &key ((:lx lx) nil) ((:ly ly) nil)
		      ((:means means) nil) ((:deviations deviations) nil)
		      ((:data data) t) ((:lines lines) nil) ((:fns fns) nil)
		      ((:file file) nil))
  (let ((item (gensym)))
    (with-open-file (outfile (format nil "/tmp/~a" item)
			     :direction :output)
      (if file (setf file (format nil "/tmp/~a.ps" item)))
      (output-a-set set list-of-planners outfile lx ly data means deviations lines file)
      (loop for (fn title) in fns
	    do (output-a-fn fn title (sub-set-stats (car set)) outfile))
      (shell (format nil "xgraph =200x400 ~A &" (namestring outfile))))))
