;;;  Contains: Many of the global variables, operator functions,
;;;            and matching code
;;;  Author:  Andy Philips (abp@goedel.arc.nasa.gov)

(defvar *OPERATORS* nil
  "Contains operator descriptions in the world")

(defvar *OP-CLASSES* nil
  "Contains operator classes")

(defvar *initial-state* nil
  "Contains the initial conjunction of predicates")

(defvar *final-state* nil
  "Contains the final conjunction of predicates")

(defvar *operator-number* 1)

(defvar *compiled-operators* (make-hash-table :test #'equal))

(defvar *PLAN* nil
  "Contains the operators involved in a plan")

(defvar *random-goal* 'comb)

(defvar *random-debug* nil) ;; If set, will be used as a seed to NEW-RANDOM

(defvar *initial-operator* nil)

(defvar *final-operator* nil)

(defvar *depth-level* 5)

(defvar *visits* 0)

(defvar *out-stream* t)

(defvar *debug* nil)

(defvar *back-halt* nil)

(defvar *node-limit* 150000)


(defun create-operator-classes (op-list)
  (let ((ops nil))
    (dolist (op op-list ops)
	    (push (make-operator-class
		   :name (car op)
		   :static (cadr (assoc 'static (cdr op)))
		   :pre (cadr (assoc 'pre (cdr op)))
		   :del (cadr (assoc 'del (cdr op)))
		   :add (cadr (assoc 'add (cdr op))))
		  ops))))


(defun create-operator-relevance-table (ops)
  (clrhash *compiled-operators*)
  (dolist (op ops)
	  (dolist (fluent (operator-class-add op))
		  (let ((tuple (gethash fluent *compiled-operators*)))
		    (if tuple (rplacd (last tuple) `(,op))
		      (setf (gethash fluent *compiled-operators*)
			    `(,op)))))))


(defun initial-state-operator (predicates)
  (make-operator
   :class (make-operator-class :name 'initial-operator :add predicates)
   :instance -1))


(defun final-state-operator (predicates)
  (let ((f (make-operator
	    :class (make-operator-class :name 'final-operator :pre predicates)
	    :instance 0)))
    (construct-circular-list f predicates (list nil))
    f))


(defun modified-delete (test-pred list-of-pred new-cl mark)
  (declare (optimize (speed 3) (safety 0)))
  (dolist (item list-of-pred)
    (declare (type pred item))
    (cond ((equal test-pred (pred-goal item))
	   (setf (pred-? item) mark)
	   (circular-add new-cl item))))
  (delete 'maybe list-of-pred :test-not #'eq :key #'pred-?))


(defun unique-instance-number ()
  (incf *operator-number*))


(defun create-operator-instance (op-class)
  (let ((op
	 (make-operator :class op-class :instance (unique-instance-number))))
    (construct-circular-list op (operator-class-pre op-class) (list nil))
    (push op *PLAN*)
    op))


(defun destroy-operator-instance (&rest y)
  (declare (ignore y))
  (let ((op (pop *PLAN*)))
    (dolist (item (operator-pre op)) (circular-delete item))
    (decf *OPERATOR-NUMBER*)))


(defun help-clear-out-marks-forward (op-set mark collect)
  (dolist (op op-set)
	  (if (not (equal (operator-mark op) mark))
	      (progn
		(setf (operator-mark op) mark)
		(if collect
		    (cons op (help-clear-out-marks-forward (operator-succs op) mark collect))
		  (help-clear-out-marks-forward (operator-succs op) mark collect))))))


(defun help-clear-out-marks-backward (op-set mark collect)
  (dolist (op op-set)
	  (if (not (equal (operator-mark op) mark))
	      (progn
		(setf (operator-mark op) mark)
		(if collect
		    (cons op (help-clear-out-marks-backward (operator-preds op) mark collect))
		  (help-clear-out-marks-backward (operator-preds op) mark collect))))))


(defun clear-out-marks (operator-set mark)
  (help-clear-out-marks-forward operator-set (gensym) nil)
  (help-clear-out-marks-forward operator-set mark nil))





;;;;; SECTION ON GENERALIZED OPERATORS
;;;;;   
;;;;;   
;;;;; Operators for blocksworld:
;;;;;   
;;;;;   LHSS: ((block ?x)(block ?z)(not-equal ?x ?z))
;;;;;   RHSD: ((on ?x table)(clear ?x)(clear ?z))
;;;;;   ADD:  ((on ?x ?z))
;;;;;   DEL:  ((on ?x table)(clear ?z))
;;;;;   
;;;;;   LHSS: ((block ?x)(block ?y)(not-equal ?x ?y))
;;;;;   LHSD: ((on ?x ?y)(clear ?x))
;;;;;   ADD:  ((on ?x table)(clear ?y))
;;;;;   DEL:  ((on ?x ?y))
;;;;;   

#| Operater specs for on type of blocks world...

(setf *operators*
      '(((move ?x table ?z)
	 (static
	  ((block ?x)(block ?z)(not-equal ?x ?z)))
	 (pre
	  ((on ?x table)(clear ?x)(clear ?z)))
	 (add
	  ((on ?x ?z)))
	 (del
	  ((on ?x table)(clear ?z))))
	((move ?x ?y table)
	 (static
	  ((block ?x)(block ?y)(not-equal ?x ?y)))
	 (pre
	  ((on ?x ?y)(clear ?x)))
	 (add
	  ((on ?x table)(clear ?y)))
	 (del
	  ((on ?x ?y))))))

(setf *initial-state* '((block a)(block b)(block c)(on a table)(clear c)
			(on b table)(on c table)(clear a)(clear b)))

(setf *final-state* '((on a b)(on b c)))

|#



#| ACHIEVEMENT-OPERATORS

This selects the operators that will achieve the goal supplied (ie.
those whose add slot contain the appropriate goal.
|#

(defun achievement-operators (goal)
  (get-all-instances
   (loop for op in *op-classes*
	 nconc (loop with bindings = nil
		     with matched? = nil
		     for spec in (operator-class-add op)
		     do (multiple-value-setq
			 (bindings matched?)
			 (pred-match spec goal nil))
		     if matched? collect (list op bindings)))))


(defun variable-substitution (set bindings)
  (loop for (var . val) in bindings
	do (setf set (subst val var set)))
  set)


;; Set1 contains variables.  Set2 contains no variables.  Bindings are those
;; variables already bound, but do not appear in Set1.
(defun get-all-binding-lists (set1 set2 original-bindings)
  (loop with bindings-lists = nil
	initially (setf bindings-lists (list original-bindings))
	for spec in set1
	if (equal (car spec) 'not-equal)
	do   (setf bindings-lists (not-equal-match-set spec bindings-lists))
	else do (setf bindings-lists (pred-match-set spec set2 bindings-lists))
	finally (return bindings-lists)))




(defun not-equal-match-set (spec bindings-lists)
  (loop for bindings in bindings-lists
	unless (equal (or (cdr (assoc (nth 1 spec) bindings)) (nth 1 spec))
		      (or (cdr (assoc (nth 2 spec) bindings)) (nth 2 spec)))
	collect bindings))


(defun pred-match-set (spec set bindings-lists)
  (loop for bindings in bindings-lists
	nconc (loop with bind = nil
		    with matched? = nil
		    initially (setf bind nil matched? nil)
		    for obj in set
		    do (multiple-value-setq
			(bind matched?)
			(pred-match spec obj bindings))
		    if (and matched? (or bind (null bindings)))
		    collect bind)))



(defun get-all-instances (op-binding-pairs)
  (loop
   with instance = nil
   for (op bindings) in op-binding-pairs
   do (setf instance
	    (variable-substitution (operator-class-static op) bindings))
      (if *debug* (format *debug* "~A~%" instance))
   nconc (loop for bindings in (get-all-binding-lists
				instance *initial-state* bindings)
	       collect (let ((name (operator-class-name op))
			     (static (operator-class-static op))
			     (pre (operator-class-pre op))
			     (add (operator-class-add op))
			     (del (operator-class-del op)))
			 (make-operator-class
			  :name (car (variable-substitution `(,name) bindings))
			  :static (variable-substitution static bindings)
			  :pre (variable-substitution pre bindings)
			  :add (variable-substitution add bindings)
			  :del (variable-substitution del bindings))))))


;; Matching code for doing functional operators
;; (pred-match '(on ?x ?y) '(on a b) '((?z . c) (?y . b)))
;; -> ((?x . a) ((?z . c) (?y . b))) , T
;;
;; (pred-match '(on ?x ?y) '(on a b) '((?y . c)))
;; -> NIL , NIL
;;

(defun pred-match (spec obj bindings)
  (loop with matched? = t
	while spec
	do (multiple-value-setq (bindings matched?)
				(value-match (car spec) (car obj) bindings))
	until (null matched?) ; something else
	do (setf spec (cdr spec) obj (cdr obj))
	finally (return (values bindings matched?))))


(defun value-match (spec-val obj-val bindings)
  (cond ((is-variable spec-val)
	 (let ((binding (assoc spec-val bindings)))
	   (cond ((null binding)
		  (values (cons (cons spec-val obj-val) bindings) t))
		 ((equal obj-val (cdr binding)) (values bindings t))
		 (t (values nil nil))))) ;fail
	((equal spec-val obj-val) (values bindings t))
	(t (values nil nil))))		;fail


(defun is-variable (item)
  (and (not (numberp item))
       (atom item)
       (equal (subseq (string item) 0 1) "?")))

