;;;  Contains: the goal ordering code for the UA unambiguous planner
;;;  Author:  Andy Philips (abp@goedel.arc.nasa.gov)


(proclaim '(optimize (speed 3) (safety 0)))


(proclaim '(special *OPERATORS* *OP-CLASSES* *initial-state*
		    *final-state* *operator-number*
		    *compiled-operators* *PLAN* *random-goal*))


(defun clear-out-all-marks (mark)
  (dolist (op *plan*) (setf (operator-mark op) mark)))


(defun help-linearize-forward (parent)
  (if (operator-mark parent) (operator-mark parent)
    (let ((newest-child nil) (newest-value 0) (temp 0))
      (dolist (child (operator-succs parent))
	      (setf temp (help-linearize-forward child))
	      (if (< newest-value temp)
		  (setf newest-value temp newest-child child)))
      (cond ((null newest-child)
	     (setf (operator-mark parent) 1.0))
	    ((null (operator-bef newest-child))
	     (setf (operator-bef newest-child) parent
		   (operator-aft parent) newest-child)
	     (setf (operator-mark parent) (1+ newest-value)))
	    (t
	     (setf (operator-aft parent) newest-child
		   (operator-bef parent) (operator-bef newest-child)
		   (operator-bef newest-child) parent
		   (operator-aft (operator-bef parent)) parent)
	     (setf (operator-mark parent)
		   (/ (+ (operator-mark (operator-bef parent))
			 newest-value) 2)))))))


(defun help-linearize-backward (child)
  (if (operator-mark child) (operator-mark child)
    (let ((newest-parent nil) (newest-value 0) (temp 0))
      (dolist (child (operator-preds child))
	      (setf temp (help-linearize-backward child))
	      (if (> newest-value temp)
		  (setf newest-value temp newest-parent child)))
      (cond ((null newest-parent)
	     (setf (operator-mark child) -1.0))
	    ((null (operator-aft newest-parent))
	     (setf (operator-aft newest-parent) child
		   (operator-bef child) newest-parent)
	     (setf (operator-mark child) (1- newest-value)))
	    (t
	     (setf (operator-bef child) newest-parent
		   (operator-aft child) (operator-aft newest-parent)
		   (operator-aft newest-parent) child
		   (operator-bef (operator-aft child)) child)
	     (setf (operator-mark child)
		   (/ (+ (operator-mark (operator-aft child))
			 newest-value) 2)))))))


(defun linearize-ua-plan (target-step)
  (dolist (op *plan*) (setf (operator-mark op) nil
			    (operator-bef  op) nil
			    (operator-aft  op) nil))
  (help-linearize-forward  target-step)
  (setf (operator-mark target-step) nil)
  (help-linearize-backward target-step))



(defun readjust-unach-goals (cl new-adds new-dels)
  (declare (optimize (speed 3) (safety 0)))
  (dolist (item new-adds)
    (declare (type pred item))
    (circular-add (first cl) item)
    (setf (pred-? item) nil))
  (dolist (item new-dels)
    (declare (type pred item))
    (circular-add (second cl) item)
    (setf (pred-? item) t)))


(defun help-ua-mark-backward (op)
  (if (numberp (operator-mark op))
      (incf (operator-mark op))
    (progn
      (setf (operator-mark op) 1)
      (dolist (temp-op (operator-preds op))
	      (help-ua-mark-backward temp-op)))))


(defun ua-mark-backward (op)
  (help-clear-out-marks-backward `(,op) (gensym) nil)
  (help-ua-mark-backward op))


(defun help-ua-mark-forward (op)
  (if (numberp (operator-mark op))
      (incf (operator-mark op))
    (progn
      (setf (operator-mark op) 1)
      (dolist (temp-op (operator-succs op))
	      (help-ua-mark-forward temp-op)))))


(defun ua-mark-forward (op)
  (help-clear-out-marks-forward `(,op) (gensym) nil)
  (help-ua-mark-forward op))


(defun help-ua-adjust-unachieved-goals-pre (op unach nec ach)
  (if (and op unach)
      (progn
	(dolist (add (operator-class-add (operator-class op)))
	  (setf unach (modified-delete add unach ach t)))
	(dolist (del (operator-class-del (operator-class op)))
	  (setf unach (modified-delete del unach nec nil)))
	(help-ua-adjust-unachieved-goals-pre (operator-bef op) unach nec ach))
    unach))


(defun help-ua-adjust-unachieved-goals-add (add new-op nec ach trash)
  (if (and new-op (car nec))
      (let ((continue t) (class (operator-class new-op)))
	(dolist (item (operator-pre new-op))
	  (cond ((not (equal (pred-goal item) add)))
		((pred-? item) (setf continue nil) (return))
		(t (circular-add ach item) (setf (pred-? item) t)
		   (push item trash) (return))))
	(if (and continue
		 (not (member add (operator-class-add class) :test #'equal))
		 (not (member add (operator-class-del class) :test #'equal)))
	    (help-ua-adjust-unachieved-goals-add
	     add (operator-aft new-op) nec ach trash)
	  trash))
    trash))


(defun help-ua-adjust-unachieved-goals-del (del new-op nec ach trash op)
  (if (and new-op (car ach))
      (let ((continue t) (class (operator-class new-op)))
	(dolist (item (operator-pre new-op))
	  (cond ((not (equal (pred-goal item) del)))
		((not (pred-? item)) (setf continue nil) (return))
		(t (circular-add nec item) (setf (pred-? item) nil)
		   (push item trash) (return))))
	(if (and continue
		 (not (member del (operator-class-add class) :test #'equal))
		 (not (member del (operator-class-del class) :test #'equal)))
	    (help-ua-adjust-unachieved-goals-del
	     del (operator-aft new-op) nec ach trash op)
	  trash))
    trash))


#| UA-ADJUST-UNACHIEVED-GOALS

This assumes that the operator is already placed within an unambiguous
plan.

|#

(defun ua-adjust-unachieved-goals (op complete-cl)
  (let ((nec (first complete-cl)) (ach (second complete-cl))
	(unach (copy-list (operator-pre op))) (new-adds nil) (new-dels nil))
    (dolist (pred (operator-pre op)) (setf (pred-? pred) 'maybe))
    (linearize-ua-plan op)
    (setf unach (help-ua-adjust-unachieved-goals-pre
		 (operator-bef op) unach nec ach))
    (dolist (item unach)
      (circular-add nec item)
      (setf (pred-? item) nil))
    (dolist (add (operator-class-add (operator-class op)))
      (setf new-adds (help-ua-adjust-unachieved-goals-add
		      add (operator-aft op) nec ach new-adds)))
    (dolist (del (operator-class-del (operator-class op)))
      (setf new-dels (help-ua-adjust-unachieved-goals-del
		      del (operator-aft op) nec ach new-dels op)))
    (values new-adds new-dels)))


(defun help-ua-get-unachieved-goals (op unach nec ach)
  (if (<= (length (operator-succs op)) (incf (operator-mark op)))
      (progn
	(dolist (add (operator-class-add (operator-class op)))
	  (setf unach (modified-delete add unach ach t)))
	(dolist (del (operator-class-del (operator-class op)))
	  (setf unach (modified-delete del unach nec nil)))
	(dolist (pred (operator-pre op)) (setf (pred-? pred) 'maybe))
	(setf unach (nconc (copy-list (operator-pre op)) unach))
	(dolist (op (operator-preds op))
	  (setf unach (help-ua-get-unachieved-goals op unach nec ach)))))
  unach)


(defun ua-get-unachieved-goals nil
  (let ((unach nil) (nec (list nil)) (ach (list nil)))
    (clear-out-all-marks 0)
    (setf unach (help-ua-get-unachieved-goals *final-operator* nil nec ach))
    (dolist (item unach)
      (setf (pred-? item) nil)
      (circular-add nec item))
    (list nec ach)))




#|(defun help-linearize-forward (parent oldest)
  (if (operator-mark parent) oldest
    (progn
      (setf (operator-mark parent) t)
      (dolist (child (operator-succs parent))
	      (setf oldest (help-linearize-forward child oldest)))
      (if oldest (setf (operator-bef oldest) parent
		       (operator-aft parent) oldest))
      parent)))


(defun help-linearize-backward (child oldest)
  (if (operator-mark child) oldest
    (progn
      (setf (operator-mark child) t)
      (dolist (parent (operator-preds child))
	      (setf oldest (help-linearize-backward parent oldest)))
      (if oldest (setf (operator-aft oldest) child
		       (operator-bef child) oldest))
      child)))


(defun linearize-ua-plan (target-step)
  (dolist (op *plan*) (setf (operator-mark op) nil
			    (operator-bef  op) nil
			    (operator-aft  op) nil))
  (help-linearize-forward  target-step nil)
  (setf (operator-mark target-step) nil)
  (help-linearize-backward target-step nil))|#


