;;;  Contains: the totally ordered planner doing iterative broadening
;;;  Author:  Andy Philips (abp@goedel.arc.nasa.gov)
;;;  Created: Mon Oct 28 10:48:31 1991

#|

(IB-TO)		Runs TO with iterative broadening.
(IB-TO-MG)	Runs TO with iterative broadening and Min Goals Heuristic.

Notes:

(NEW-RANDOM x) is the same as (RANDOM X)

This system requires the LOOP macro.

Try loading problems from the P/ directory.  Try 'problems.lisp' and
call the function '(PROBLEM 1)'.

These files work best when compiled.

(setf *graphics-on* nil) when not in ALLEGRO with common windows.

|#

(proclaim '(optimize (speed 3) (safety 0)))


(proclaim '(special *initial-operator* *final-operator* *depth-level*
		    *visits* *out-stream* *debug* *back-halt* *node-limit*))


#| IB-TO-EXTEND-FAST

Given a plan and a goal set, this will find an operator that achieves
one of the goals and will then place this operator in the plan in the
appropriate location, recursing on the new goal-set.

This computes the new goal set based on the operator added.

|#



(defun ib-to-extend-fast (complete-cl c limit &optional (level 0))
  (cond
   ((null (caar complete-cl)) t)
   ((not (< level *depth-level*)) nil)
   ((> *visits* limit) nil)
   (t (let ((goal-to-achieve (select-goal (first complete-cl) level)) (op nil)
	    (operators nil) (heuristic nil))
	(if *debug* (format *debug* "G: ~A~%" (pred-goal goal-to-achieve)))
	(setf operators (achievement-operators (pred-goal goal-to-achieve)))
	(if (eql *random-goal* 'random) (shuffle-list operators))
	(dolist (op-class operators nil)
	  (setf op (create-operator-instance op-class))
	  (if *debug* (format *debug* "OP: ~A~%"
			      (operator-class-name (operator-class op))))
	  (loop with position = (pred-op goal-to-achieve)
		while (setf position (car (operator-preds position))) do
		(push `(,op-class ,position) heuristic)
		until (member (pred-goal goal-to-achieve)
			      (operator-class-del (operator-class position))
			      :test #'equal))
	  (destroy-operator-instance))
	(loop with tuple = nil
	      with position = nil
	      with count = 0
	      while (and (<= (incf count) c) (setf tuple (pop heuristic))) do
	      (setf op (create-operator-instance (car tuple))
		    position (nth 1 tuple))
	      (if *debug* (format *debug* "OP: ~A~%"
				  (operator-class-name (operator-class op))))
	      (place-operator-after op position)
	      (graphics-add-step *plan* level)
	      (multiple-value-bind
	       (n-adds n-dels) (adjust-unachieved-goals op complete-cl)
	       (if *debug* (format *debug* "~A ~A~%" *plan* position))
	       (incf *visits*)
	       (if (ib-to-extend-fast complete-cl c limit (1+ level))
		   (return t))
	       (remove-operator-from-plan op)
	       (readjust-unach-goals complete-cl n-adds n-dels)
	       (graphics-del-step *plan* level))
	      (destroy-operator-instance))))))


(defun ib-mod2-to-extend-fast (complete-cl c limit &optional (level 0))
  (cond
   ((null (caar complete-cl)) t)
   ((not (< level *depth-level*)) nil)
   ((> *visits* limit) nil)
   (t (let ((goal-to-achieve (select-goal (first complete-cl) level)) (op nil)
	    (operators nil) (heuristic nil))
	(if *debug* (format *debug* "G: ~A~%" (pred-goal goal-to-achieve)))
	(setf operators (achievement-operators (pred-goal goal-to-achieve)))
	(if (eql *random-goal* 'random) (shuffle-list operators))
	(dolist (op-class operators nil)
	  (setf op (create-operator-instance op-class))
	  (if *debug* (format *debug* "OP: ~A~%"
			      (operator-class-name (operator-class op))))
	  (loop with position = (pred-op goal-to-achieve)
		while (setf position (car (operator-preds position))) do
		(place-operator-after op position)
		(multiple-value-bind
		 (n-adds n-dels) (adjust-unachieved-goals op complete-cl)
		 (let ((h (- (length n-dels) (length n-adds))))
		   (dolist (pre (operator-pre op))
			   (if (not (pred-? pre)) (incf h)))
		   (setf heuristic
			 (insert-ordering h `(,op-class ,position) heuristic)))
		 (remove-operator-from-plan op)
		 (readjust-unach-goals complete-cl n-adds n-dels))
		until (member (pred-goal goal-to-achieve)
			      (operator-class-del (operator-class position))
			      :test #'equal))
	  (destroy-operator-instance))
	(setf heuristic (sort heuristic #'< :key #'car))
	(loop with tuple = nil
	      with position = nil
	      with count = 0
	      while (and (<= (incf count) c)
			 (setf tuple (extract-ordering heuristic))) do
	      (setf op (create-operator-instance (car tuple))
		    position (nth 1 tuple))
	      (if *debug* (format *debug* "OP: ~A~%"
				  (operator-class-name (operator-class op))))
	      (place-operator-after op position)
	      (graphics-add-step *plan* level)
	      (multiple-value-bind
	       (n-adds n-dels) (adjust-unachieved-goals op complete-cl)
	       (if *debug* (format *debug* "~A ~A~%" *plan* position))
	       (incf *visits*)
	       (if (ib-mod2-to-extend-fast complete-cl c limit (1+ level))
		   (return t))
	       (remove-operator-from-plan op)
	       (readjust-unach-goals complete-cl n-adds n-dels)
	       (graphics-del-step *plan* level))
	      (destroy-operator-instance))))))


(defun ib-to nil
  (let* ((t1 0) (result nil) (t2 0) (final-c 0))
    (setf t1 (get-internal-run-time))
    (loop with nodes = 0
	  with limit = *node-limit*
	  with c = 0
	  do (to)
	     (incf c)
	     (setf result (ib-to-extend-fast (get-unachieved-goals) c limit))
	     (incf nodes *visits*)
	     (decf limit *visits*)
	  until (or result (< limit 1))
	  finally (setf *visits* nodes final-c c))
    (setf t2 (get-internal-run-time))
    (format *out-stream* "Number of calls: ~A~%" *visits*)
    (dolist (i *plan*) (format *out-stream* "~A~%" i))
    (show-to-plan)
    `(,result ,*visits* ,(/ (- t2 t1) internal-time-units-per-second 1.0)
      ,final-c)))



(defun ib-to-mg nil
  (let* ((t1 0) (result nil) (t2 0) (final-c 0))
    (setf t1 (get-internal-run-time))
    (loop with nodes = 0
	  with limit = *node-limit*
	  with c = 0
	  do (to)
	     (incf c)
	     (setf result
		   (ib-mod2-to-extend-fast (get-unachieved-goals) c limit))
	     (incf nodes *visits*)
	     (decf limit *visits*)
	  until (or result (< limit 1))
	  finally (setf *visits* nodes final-c c))
    (setf t2 (get-internal-run-time))
    (format *out-stream* "Number of calls: ~A~%" *visits*)
    (dolist (i *plan*) (format *out-stream* "~A~%" i))
    (show-to-plan)
    `(,result ,*visits* ,(/ (- t2 t1) internal-time-units-per-second 1.0)
      ,final-c)))
