;;;  Contains: the unambiguous planner doing iterative broadening
;;;  Author:  Andy Philips (abp@goedel.arc.nasa.gov)
;;;  Created: Mon Oct 28 10:46:32 1991

#|

Func	Algorithm

(IB-UA)		Runs UA with iterative broadening.
(IB-UA-MG)	Runs UA with iterative broadening and Min Goals Heuristic.


None of these functions take any arguments.

Notes:

(NEW-RANDOM x) is the same as (RANDOM X)

This system requires the LOOP macro.

Try loading problems from the P/ directory.  Try 'problems.lisp' and
call the function '(PROBLEM 1)'.

These files work best when compiled.

(setf *graphics-on* nil) when not in ALLEGRO with common windows.

|#

(proclaim '(optimize (speed 3) (safety 0)))


(proclaim '(special *initial-operator* *final-operator* *depth-level*
		    *visits* *out-stream* *debug* *back-halt* *node-limit*))



(defun ib-ua-extend-fast (complete-cl c limit &optional (level 0))
  (cond
   ((null (caar complete-cl)) t)
   ((not (< level *depth-level*)) nil)
   ((> *visits* limit) nil)
   (t (let ((pre (select-goal (first complete-cl) level)) (op nil)
	    (deleter nil) (heuristic nil))
	(if *debug* (format *debug* "G: ~A~%" (pred-goal pre)))
	(dolist (op-class (achievement-operators (pred-goal pre)))
	  (setf op (create-operator-instance op-class)
		deleter (find-deleter pre))
	  (add-ordering deleter op)
	  (add-ordering op (pred-op pre))
	  (loop with state = (disambiguate-setup op #'always-after)
		with flag  = (if (nth 1 state) t)
		with stack = nil
		while (setf state (disambiguate-forward op state)) do
		(setf stack (if flag (copy-stack-from-state state) nil))
		(push (list op-class stack deleter) heuristic)
		(if flag (setf state (disambiguate-backtrack op state)))
		while (and state flag))
	  (remove-ordering op (pred-op pre))
	  (remove-ordering deleter op)
	  (destroy-operator-instance))
	(loop with tuple = nil
	      with stack = nil
	      with count = 0
	      while (and (<= (incf count) c) (setf tuple (pop heuristic))) do
	      (setf op (create-operator-instance (nth 0 tuple))
		    stack (nth 1 tuple)
		    deleter (nth 2 tuple))
	      (add-ordering deleter op)
	      (add-ordering op (pred-op pre))
	      (if *debug* (format *debug* "OP: ~A~%"
				  (operator-class-name (operator-class op))))
	      (reinstate op stack)
	      (graphics-add-step *plan* level)
	      (multiple-value-bind
	       (n-adds n-dels) (ua-adjust-unachieved-goals op complete-cl)
	       (if *debug* (progn (format t "U: ~D, A: ~D, " (length n-dels)
					  (length n-adds)) (ua-order-plan)))
	       (incf *visits*)
	       (if (ib-ua-extend-fast complete-cl c limit (1+ level))
		   (return t))
	       (deinstate op stack)
	       (readjust-unach-goals complete-cl n-adds n-dels)
	       (graphics-del-step *plan* level))
	      (remove-ordering op (pred-op pre))
	      (remove-ordering deleter op)
	      (destroy-operator-instance))))))


(defun ib-mod3-ua-extend-fast (complete-cl c limit &optional (level 0))
  (cond
   ((null (caar complete-cl)) t)
   ((not (< level *depth-level*)) nil)
   ((> *visits* limit) nil)
   (t (let ((pre (select-goal (first complete-cl) level)) (op nil)
	    (deleter nil) (heuristic nil))
	(if *debug* (format *debug* "G: ~A~%" (pred-goal pre)))
	(dolist (op-class (achievement-operators (pred-goal pre)))
	  (setf op (create-operator-instance op-class)
		deleter (find-deleter pre))
	  (add-ordering deleter op)
	  (add-ordering op (pred-op pre))
	  (loop with state = (disambiguate-setup op #'always-after)
		with flag  = (if (nth 1 state) t)
		with stack = nil
		with stuff = nil
		while (setf state (disambiguate-forward op state)) do
		(multiple-value-bind
		 (n-adds n-dels)
		 (ua-adjust-unachieved-goals op complete-cl)
		 (let ((h (- (length n-dels) (length n-adds))))
		   (loop for pre in (operator-pre op)
			 if (not (pred-? pre)) do (incf h))
		   (setf stack (if flag (copy-stack-from-state state) nil)
			 stuff (list op-class stack deleter))
		   (setf heuristic (insert-ordering h stuff heuristic)))
		 (if flag (setf state (disambiguate-backtrack op state)))
		 (readjust-unach-goals complete-cl n-adds n-dels))
		while (and state flag))
	  (remove-ordering op (pred-op pre))
	  (remove-ordering deleter op)
	  (destroy-operator-instance))
	(setf heuristic (sort heuristic #'< :key #'car))
	(loop with tuple = nil
	      with stack = nil
	      with count = 0
	      while (and (<= (incf count) c)
			 (setf tuple (extract-ordering heuristic))) do
	      (setf op (create-operator-instance (nth 0 tuple))
		    stack (nth 1 tuple)
		    deleter (nth 2 tuple))
	      (add-ordering deleter op)
	      (add-ordering op (pred-op pre))
	      (if *debug* (format *debug* "OP: ~A~%"
				  (operator-class-name (operator-class op))))
	      (reinstate op stack)
	      (graphics-add-step *plan* level)
	      (multiple-value-bind
	       (n-adds n-dels) (ua-adjust-unachieved-goals op complete-cl)
	       (if *debug* (progn (format t "U: ~D, A: ~D, " (length n-dels)
					  (length n-adds)) (ua-order-plan)))
	       (incf *visits*)
	       (if (ib-mod3-ua-extend-fast complete-cl c limit (1+ level))
		   (return t))
	       (deinstate op stack)
	       (readjust-unach-goals complete-cl n-adds n-dels)
	       (graphics-del-step *plan* level))
	      (remove-ordering op (pred-op pre))
	      (remove-ordering deleter op)
	      (destroy-operator-instance))))))


(defun ib-ua nil
  (let* ((t1 0) (result nil) (t2 0) (final-c 0))
    (setf t1 (get-internal-run-time))
    (loop with nodes = 0
	  with limit = *node-limit*
	  with c = 0
	  do (ua)
	     (incf c)
	     (setf result
		   (ib-ua-extend-fast (ua-get-unachieved-goals) c limit))
	     (incf nodes *visits*)
	     (decf limit *visits*)
	  until (or result (< limit 1))
	  finally (setf *visits* nodes final-c c))
    (setf t2 (get-internal-run-time))
    (format *out-stream* "Number of calls: ~A~%" *visits*)
    (dolist (i *plan*) (format *out-stream* "~A~%" i))
    (show-to-plan)
    `(,result ,*visits* ,(/ (- t2 t1) internal-time-units-per-second 1.0)
      ,final-c)))



(defun ib-ua-mg nil
  (let* ((t1 0) (result nil) (t2 0) (final-c 0))
    (setf t1 (get-internal-run-time))
    (loop with nodes = 0
	  with limit = *node-limit*
	  with c = 0
	  do (ua)
	     (incf c)
	     (setf result
		   (ib-mod3-ua-extend-fast (ua-get-unachieved-goals) c limit))
	     (incf nodes *visits*)
	     (decf limit *visits*)
	  until (or result (< limit 1))
	  finally (setf *visits* nodes final-c c))
    (setf t2 (get-internal-run-time))
    (format *out-stream* "Number of calls: ~A~%" *visits*)
    (dolist (i *plan*) (format *out-stream* "~A~%" i))
    (show-to-plan)
    `(,result ,*visits* ,(/ (- t2 t1) internal-time-units-per-second 1.0)
      ,final-c)))
