;;;  Contains: the totally ordered planner, TO
;;;  Author:  Andy Philips (abp@goedel.arc.nasa.gov)

#|
Commands to call from within this system are:

(TOF)	Runs the TO algorithm (Fast goal update version).
(MTOF)	Fast version with heuristic search method that checks all orderings
	of a new operator and selects the best ordering.
(FTOF)	Fast version with Minimum Goals Heuristic.

Notes:

(NEW-RANDOM x) is the same as (RANDOM X)

This system requires the LOOP macro.

Try loading problems from the P/ directory.  Try 'problems.lisp' and
call the function '(PROBLEM 1)'.

These files work best when compiled.

(setf *graphics-on* nil) when not in ALLEGRO with common windows.

|#

(proclaim '(optimize (speed 3) (safety 0)))


(proclaim '(special *initial-operator* *final-operator* *depth-level*
		    *visits* *out-stream* *debug* *back-halt* *node-limit*))


(defun to ()
  (setq *num-sols* 0) (setq *num-fails* 0)
  (format *out-stream* "Planning to depth level ~A~%" *depth-level*)
  (format *out-stream* "Goal selection will be ~A~%"  *random-goal*)
  (if *random-debug*
      (progn
	(format (or *out-stream* *debug* t) "RANDOM SEED being used!!!~%")
	(setf *new-random* (copy-seq *random-debug*))))
  (setf *operator-number* 0 *visits* 0)
  (setf *initial-operator* (initial-state-operator *initial-state*)
	*final-operator* (final-state-operator *final-state*))
  (setf *OP-CLASSES* (create-operator-classes *operators*))
;  (create-operator-relevance-table *OP-CLASSES*)
  (push *final-operator* (operator-succs *initial-operator*))
  (push *initial-operator* (operator-preds *final-operator*))
  (setf *PLAN* `(,*final-operator* ,*initial-operator*))
  (init-graphics *depth-level* *plan* "TO")
  *plan*)


(defun place-operator-after (op back)
  (let ((ahead (car (operator-succs back))))
    (push op (operator-succs back))
    (push op (operator-preds ahead))
    (push back (operator-preds op))
    (push ahead (operator-succs op))))


(defun remove-operator-from-plan (op)
  (if (operator-preds op) (pop (operator-succs (car (operator-preds op)))))
  (if (operator-succs op) (pop (operator-preds (car (operator-succs op)))))
  (setf (operator-succs op) nil (operator-preds op) nil)
  (dolist (item (operator-pre op)) (circular-delete item)))


#| TO-EXTEND-SLOW

Given a plan and a goal set, this will find an operator that achieves
one of the goals and will then place this operator in the plan in the
appropriate location, recursing on the new goal-set.

It will recompute the new goal set at each step.

|#

(defun to-extend-slow (complete-cl &optional (level 0))
  (cond ((null (caar complete-cl)) t)
	((not (< level *depth-level*)) nil)
	((> *visits* *node-limit*) nil)
	(t
	 (let ((goal-to-achieve
		(select-goal (first complete-cl) level)))
	   (loop with operator = nil
		 with operators = (achievement-operators
				   (pred-goal goal-to-achieve))
		 initially (if (eql *random-goal* 'random)
			       (shuffle-list operators))
		 for operator-class in operators
		 do (setf operator (create-operator-instance operator-class))
		 if (loop with position = (pred-op goal-to-achieve)
			  do (setf position (car (operator-preds position)))
			  while position
		          do (place-operator-after operator position)
			     (incf *visits*)
			  if (to-extend-slow (get-unachieved-goals) (1+ level))
			  do (return t)
			  else do (remove-operator-from-plan operator)
			  until (member (pred-goal goal-to-achieve)
					(operator-class-del operator-class)
					:test #'equal))
		 do (return t)
		 else do (destroy-operator-instance))))))


#| TO-EXTEND-FAST

Given a plan and a goal set, this will find an operator that achieves
one of the goals and will then place this operator in the plan in the
appropriate location, recursing on the new goal-set.

This computes the new goal set based on the operator added.

|#

(defun to-extend-fast (complete-cl &optional (level 0))
  (cond
   ((null (caar complete-cl))
    (setf *num-sols* (1+ *num-sols*))
    nil)
   ((not (< level *depth-level*)) 
    (setf *num-fails* (1+ *num-fails*))
    nil)
   ((> *visits* *node-limit*) nil)
   (t (let ((goal-to-achieve (select-goal (first complete-cl) level))
	    (new-adds nil) (new-dels nil))
	(if *debug* (format *debug* "G: ~A~%" (pred-goal goal-to-achieve)))
	(loop with operator = nil
	      with operators = (achievement-operators
				(pred-goal goal-to-achieve))
	      initially (if (eql *random-goal* 'random) (shuffle-list operators))
	      for operator-class in operators
	      do (setf operator (create-operator-instance operator-class))
		 (if *debug* (format *debug* "OP: ~A~%"
			 (operator-class-name (operator-class operator))))
	      if (loop with position = (pred-op goal-to-achieve)
		       do (setf position (car (operator-preds position)))
		       while position
		       do (place-operator-after operator position)
			  (graphics-add-step *plan* level)
		          (multiple-value-setq
			   (new-adds new-dels)
			   (adjust-unachieved-goals operator complete-cl))
			  (if *debug*
			      (format *debug* "~A ~A~%" *plan* position))
			  (incf *visits*)
			  (cond ((to-extend-fast complete-cl (1+ level))
				 (return t))
				(t (remove-operator-from-plan operator)
				   (readjust-unach-goals
				    complete-cl new-adds new-dels)))
			  (if *back-halt* (break))
			  (graphics-del-step *plan* level)
		       until (member (pred-goal goal-to-achieve)
				     (operator-class-del
				      (operator-class position))
				     :test #'equal))
	      do (return t)
	      else do (destroy-operator-instance))))))


(defun mod-to-extend-fast (complete-cl &optional (level 0))
  (cond
   ((null (caar complete-cl)) t)
   ((not (< level *depth-level*)) nil)
   ((> *visits* *node-limit*) nil)
   (t (let ((goal-to-achieve (select-goal (first complete-cl) level)) (op nil)
	    (operators nil) (heuristic nil) (flag nil))
	(if *debug* (format *debug* "G: ~A~%" (pred-goal goal-to-achieve)))
	(setf operators (achievement-operators (pred-goal goal-to-achieve)))
	(if (eql *random-goal* 'random) (shuffle-list operators))
	(dolist (op-class operators nil)
	  (setf op (create-operator-instance op-class) heuristic nil)
	  (if *debug* (format *debug* "OP: ~A~%"
			      (operator-class-name (operator-class op))))
	  (loop with position = (pred-op goal-to-achieve)
		while (setf position (car (operator-preds position))) do
		(place-operator-after op position)
		(multiple-value-bind
		 (n-adds n-dels) (adjust-unachieved-goals op complete-cl)
		 (let ((h (- (length n-dels) (length n-adds))))
		   (dolist (pre (operator-pre op))
			   (if (not (pred-? pre)) (incf h)))
		   (setf heuristic (insert-ordering h position heuristic)
			 flag t))
		 (remove-operator-from-plan op)
		 (readjust-unach-goals complete-cl n-adds n-dels))
		until (member (pred-goal goal-to-achieve)
			      (operator-class-del (operator-class position))
			      :test #'equal))
	  (setf heuristic (sort heuristic #'< :key #'car))
	  (if (loop
	       with position = nil
	       while (setf position (extract-ordering heuristic)) do
	       (place-operator-after op position)
	       (graphics-add-step *plan* level)
	       (multiple-value-bind
		(n-adds n-dels) (adjust-unachieved-goals op complete-cl)
		(if *debug* (format *debug* "~A ~A~%" *plan* position))
		(incf *visits*)
		(if (mod-to-extend-fast complete-cl (1+ level)) (return t))
		(remove-operator-from-plan op)
		(readjust-unach-goals complete-cl n-adds n-dels)
		(graphics-del-step *plan* level))
	       until (member (pred-goal goal-to-achieve)
			     (operator-class-del (operator-class position))
			     :test #'equal)) (return t))
	  (destroy-operator-instance))))))


(defun mod2-to-extend-fast (complete-cl &optional (level 0))
  (cond
   ((null (caar complete-cl)) t)
   ((not (< level *depth-level*)) nil)
   ((> *visits* *node-limit*) nil)
   (t (let ((goal-to-achieve (select-goal (first complete-cl) level)) (op nil)
	    (operators nil) (heuristic nil))
	(if *debug* (format *debug* "G: ~A~%" (pred-goal goal-to-achieve)))
	(setf operators (achievement-operators (pred-goal goal-to-achieve)))
	(if (eql *random-goal* 'random) (shuffle-list operators))
	(dolist (op-class operators nil)
	  (setf op (create-operator-instance op-class))
	  (if *debug* (format *debug* "OP: ~A~%"
			      (operator-class-name (operator-class op))))
	  (loop with position = (pred-op goal-to-achieve)
		while (setf position (car (operator-preds position))) do
		(place-operator-after op position)
		(multiple-value-bind
		 (n-adds n-dels) (adjust-unachieved-goals op complete-cl)
		 (let ((h (- (length n-dels) (length n-adds))))
		   (dolist (pre (operator-pre op))
			   (if (not (pred-? pre)) (incf h)))
		   (setf heuristic
			 (insert-ordering h `(,op-class ,position) heuristic)))
		 (remove-operator-from-plan op)
		 (readjust-unach-goals complete-cl n-adds n-dels))
		until (member (pred-goal goal-to-achieve)
			      (operator-class-del (operator-class position))
			      :test #'equal))
	  (destroy-operator-instance))
	(setf heuristic (sort heuristic #'< :key #'car))
	(loop with tuple = nil
	      with position = nil
	      while (setf tuple (extract-ordering heuristic)) do
	      (setf op (create-operator-instance (car tuple))
		    position (nth 1 tuple))
	      (if *debug* (format *debug* "OP: ~A~%"
				  (operator-class-name (operator-class op))))
	      (place-operator-after op position)
	      (graphics-add-step *plan* level)
	      (multiple-value-bind
	       (n-adds n-dels) (adjust-unachieved-goals op complete-cl)
	       (if *debug* (format *debug* "~A ~A~%" *plan* position))
	       (incf *visits*)
	       (if (mod2-to-extend-fast complete-cl (1+ level)) (return t))
	       (remove-operator-from-plan op)
	       (readjust-unach-goals complete-cl n-adds n-dels)
	       (graphics-del-step *plan* level))
	      (destroy-operator-instance))))))


(defun show-to-plan nil
  (loop with op = *initial-operator*
	while op
	collect op
	do (setf op (car (operator-succs op)))))


(defun tos nil
  (to)
  (let ((result (to-extend-slow (get-unachieved-goals))))
    (format *out-stream* "Number of calls: ~A~%" *visits*)
    (dolist (i *plan*) (format t "~A~%" i))
    (show-to-plan)
    `(,result ,*visits*)))


(defun tof nil
  (to)
  (let* ((t1 (get-internal-run-time))
	 (result (to-extend-fast (get-unachieved-goals)))
	 (t2 (get-internal-run-time)))
    (format *out-stream* "Number of calls: ~A~%" *visits*)
    (dolist (i *plan*) (format *out-stream* "~A~%" i))
    (show-to-plan)
    (pprint *num-sols*)
    (pprint *num-fails*)
    `(,result ,*visits* ,(/ (- t2 t1) internal-time-units-per-second 1.0)
	      	      ,*num-sols* ,*num-fails* )))


(defun btof nil
  (let* ((original *depth-level*)
	 (result (loop for depth from 1 upto original
		       do (setf *depth-level* depth)
		       (to)
		       if (to-extend-fast (get-unachieved-goals))
		       do (return t))))
    (format *out-stream* "Number of calls: ~A~%" *visits*)
    (if result (dolist (i *plan* t) (format *out-stream* "~A~%" i)))
    `(,result ,*visits*)))


(defun mtof nil
  (to)
  (let ((t1 (get-internal-run-time))
	(result (mod-to-extend-fast (get-unachieved-goals)))
	(t2 (get-internal-run-time)))
    (format *out-stream* "Number of calls: ~A~%" *visits*)
    (dolist (i *plan*) (format *out-stream* "~A~%" i))
    `(,result ,*visits* ,(/ (- t2 t1) internal-time-units-per-second 1.0))))


(defun ftof nil
  (to)
  (let ((t1 (get-internal-run-time))
	(result (mod2-to-extend-fast (get-unachieved-goals)))
	(t2 (get-internal-run-time)))
    (format *out-stream* "Number of calls: ~A~%" *visits*)
    (dolist (i *plan*) (format *out-stream* "~A~%" i))
    `(,result ,*visits* ,(/ (- t2 t1) internal-time-units-per-second 1.0))))
