;;;  Contains: the unambiguous planner, UA
;;;  Author:  Andy Philips (abp@goedel.arc.nasa.gov)


#|

Func	Algorithm

UAF	Unambiguous Planner with a fast goal update scheme

PUAF	UAF with the goal minimization heuristic used to select operator
	orderings within incremental diasmbiguation (operator selection
	is made with uniform probability).

CUAF	UAF with the goal minimization heuristic used to select operator
	orderings within full generation diasmbiguation (operator selection
	is made with uniform probability).

FUAF	UAF with the goal minimization heuristic used to select operators
	and their orderings within full generation diasmbiguation.


None of these functions take any arguments.

Notes:

(NEW-RANDOM x) is the same as (RANDOM X)

This system requires the LOOP macro.

Try loading problems from the P/ directory.  Try 'problems.lisp' and
call the function '(PROBLEM 1)'.

These files work best when compiled.

(setf *graphics-on* nil) when not in ALLEGRO with common windows.

|#

(proclaim '(optimize (speed 3) (safety 0)))


(proclaim '(special *initial-operator* *final-operator* *depth-level*
		    *visits* *out-stream* *debug* *back-halt*
		    *node-limit* *random-debug*))


(defun ua ()
  (setq *num-sols* 0) (setq *num-fails* 0)
  (format *out-stream* "Planning to depth level ~A~%" *depth-level*)
  (format *out-stream* "Goal selection will be ~A~%"  *random-goal*)
  (if *random-debug*
      (progn
	(format (or *out-stream* *debug* t) "RANDOM SEED being used!!!~%")
	(setf *new-random* (copy-seq *random-debug*))))
  (setf *operator-number* 0 *visits* 0)
  (setf *initial-operator* (initial-state-operator *initial-state*)
	*final-operator* (final-state-operator *final-state*))
  (setf *OP-CLASSES* (create-operator-classes *operators*))
;  (create-operator-relevance-table *OP-CLASSES*)
  (push *final-operator* (operator-succs *initial-operator*))
  (push *initial-operator* (operator-preds *final-operator*))
  (setf *PLAN* `(,*final-operator* ,*initial-operator*))
  (init-graphics *depth-level* *plan* "UA")
  *plan*)



(defun help-find-deleter (op goal mark orig-op)
  (let ((deleter nil))
    (if (not (eq mark (operator-mark op)))
	(let ((temp nil))
	  (setf (operator-mark op) mark)
	  (dolist (pred (operator-preds op))
		  (if (setf temp (help-find-deleter pred goal mark orig-op))
		      (setf deleter temp)))
	  (if (and (not (eql orig-op op))
		   (member goal (operator-class-del (operator-class op))
			   :test #'equal))
	      op deleter)))))


(defun find-deleter (pre)
  (let ((deleter (help-find-deleter
		  (pred-op pre) (pred-goal pre) (gensym) (pred-op pre))))
    (if (null deleter) *initial-operator* deleter)))


#| These two functions expect that they will be called in the proper
   sequence, that is when the proper backtracking point is reached. |#


(defun add-ordering (pred succ)
  "pred < succ"
  (push pred (operator-preds succ))
  (push succ (operator-succs pred)))


(defun remove-ordering (pred succ)
  "pred < succ"
  (pop (operator-preds succ))
  (pop (operator-succs pred)))


(defun ambiguous-p (op1 op2)
  (let ((oc1 (operator-class op1)) (oc2 (operator-class op2)))
    (or (intersection
	 (operator-class-pre oc1) (operator-class-add oc2) :test #'equal)
	(intersection
	 (operator-class-pre oc1) (operator-class-del oc2) :test #'equal)
	(intersection
	 (operator-class-pre oc2) (operator-class-add oc1) :test #'equal)
	(intersection
	 (operator-class-pre oc2) (operator-class-del oc1) :test #'equal))))


(defun dis-clear-forward (op-set)
  (loop for op in op-set
	if (not (car (operator-disam op)))
	do (setf (car (operator-disam op)) t)
	and collect op
	and nconc (dis-clear-forward (operator-succs op))))


(defun dis-clear-backward (op-set)
  (loop for op in op-set
	if (not (car (operator-disam op)))
	do (setf (car (operator-disam op)) t)
	and collect op
	and nconc (dis-clear-backward (operator-preds op))))


(defun disambiguate-setup (op function)
  (dolist (op *plan*) (push nil (operator-disam op)))
  (dis-clear-forward  (operator-succs op))
  (dis-clear-backward (operator-preds op))
  (setf (car (operator-disam op)) t)
  (list nil (loop for target in *PLAN*
		  if (and (not (car (operator-disam target)))
			  (ambiguous-p target op))
		  collect target)
	function))


(defun propogate-after-ordering (ambiguous-op op state)
  (add-ordering ambiguous-op op)
  (let ((top (car (nth 0 state))))
    (push 'after (nth 1 top))
    (setf (nth 2 top) (dis-clear-backward `(,ambiguous-op)))
    (setf (nth 3 top) (remove-if-not #'car (nth 1 state)
				     :key #'operator-disam))
    (setf (nth 1 state) (delete-if #'car (nth 1 state)
				   :key #'operator-disam))))


(defun propogate-before-ordering (ambiguous-op op state)
  (add-ordering op ambiguous-op)
  (let ((top (car (nth 0 state))))
    (push 'before (nth 1 top))
    (setf (nth 2 top) (dis-clear-forward `(,ambiguous-op)))
    (setf (nth 3 top) (remove-if-not #'car (nth 1 state)
				     :key #'operator-disam))
    (setf (nth 1 state) (delete-if #'car (nth 1 state)
				   :key #'operator-disam))))


(defun disambiguate-forward (op state)
  (let ((ambiguous-op (car (nth 1 state))))
    (cond (ambiguous-op
	   (push (list ambiguous-op nil nil nil) (nth 0 state))
	   (if (equal (funcall (nth 2 state) op ambiguous-op) 'before)
	       (propogate-before-ordering ambiguous-op op state)
	       (propogate-after-ordering  ambiguous-op op state))
	   (disambiguate-forward op state))
	  (t state))))


(defun disambiguate-backtrack (op state)
  (let ((top (car (nth 0 state))))
    (case (car (nth 1 top))
	  (after (remove-ordering (car top) op))
	  (before (remove-ordering op (car top))))
    (dolist (temp-op (nth 2 top))
	    (setf (car (operator-disam temp-op)) nil))
    (setf (nth 1 state) (nconc (nth 3 top) (nth 1 state)))
    (cond ((null (nth 0 state)) 
	   (dolist (op *plan* nil) (pop (operator-disam op))))
	  ((= (length (nth 1 top)) 2)
	   (pop (nth 0 state))
	   (disambiguate-backtrack op state))
	  ((eql (car (nth 1 top)) 'after)
	   (propogate-before-ordering (car top) op state)
	   state)
	  (t
	   (propogate-after-ordering (car top) op state)
	   state))))


(defun after-or-random (&rest y)
  (declare (ignore y))
  (if (and (equal *random-goal* 'random) (zerop (new-random 2)))
      'before 'after))


(defun always-after (&rest y)
  (declare (ignore y))
  'after)


;; returns 'before if |DEL(N) x PRE(A)| < |PRE(N) x DEL(A)|, or
;; if putting the new step before the ambiguous step causes fewer deletes,
;; otherwise it returns 'after
(defun determine-order (new-step ambig-step)
  (if (< (length (intersection (operator-class-del (operator-class new-step))
			       (loop for p in (operator-pre ambig-step)
				     if (eq (pred-? p) t)
				     collect (pred-goal p))
			       :test #'equal))
	 (length (intersection
		  (operator-class-pre (operator-class new-step))
		  (operator-class-del (operator-class ambig-step))
		  :test #'equal)))
      'before
      'after))


;; To make UA-EXTEND-SLOW have random step placement instead of this weird
;; front to back placement, replace the function #'ALWAYS-AFTER with the
;; function #'AFTER-OR-RANDOM.

(defun ua-extend-slow (complete-cl &optional (level 0))
  (cond
   ((null (caar complete-cl)) t)
   ((not (< level *depth-level*)) nil)
   ((> *visits* *node-limit*) nil)
   (t (let ((pre (select-goal (first complete-cl) level)))
	(loop with operator = nil
	      with deleter  = nil
	      with operators = (achievement-operators (pred-goal pre))
	      initially (if (eql *random-goal* 'random)
			    (shuffle-list operators))
	      for operator-class in operators
	      do (setf operator (create-operator-instance operator-class))
	         (setf deleter (find-deleter pre))
		 (add-ordering deleter operator)
		 (add-ordering operator (pred-op pre))
	      if (loop with state = (disambiguate-setup
				     operator #'ALWAYS-AFTER)
		       do (setf state (disambiguate-forward operator state))
		          (incf *visits*)
		       if (ua-extend-slow (ua-get-unachieved-goals) (1+ level))
		       do (return t)
		       else do (setf state (disambiguate-backtrack operator state))
		       until (null state))
	      do (return t)
	      else do (progn (remove-ordering operator (pred-op pre))
			     (remove-ordering deleter operator)
			     (destroy-operator-instance)))))))


;; To make UA-EXTEND-FAST have random step placement instead of this weird
;; front to back placement, replace the function #'ALWAYS-AFTER with the
;; function #'AFTER-OR-RANDOM.

(defun ua-extend-fast (complete-cl &optional (level 0))
;;  (if (and (eq level 5) (zerop (new-random 10))) (break))
  (cond
   ((null (caar complete-cl)) 
    (setf *num-sols* (1+ *num-sols*))
    nil)
   ((not (< level *depth-level*)) 
    (setf *num-fails* (1+ *num-fails*))
    nil)
   ((> *visits* *node-limit*) nil)
   (t (let ((pre (select-goal (first complete-cl) level))
	    (new-adds nil) (new-dels nil))
	(if *debug* (format *debug* "G: ~A~%" (pred-goal pre)))
	(loop with operator = nil
	      with deleter  = nil
	      with operators = (achievement-operators (pred-goal pre))
	      initially (if (eql *random-goal* 'random)
			    (shuffle-list operators))
	      for operator-class in operators
	      do (setf operator (create-operator-instance operator-class))
	         (setf deleter (find-deleter pre))
		 (add-ordering deleter operator)
		 (add-ordering operator (pred-op pre))
		 (if *debug* (format *debug* "OP: ~A~%"
			 (operator-class-name (operator-class operator))))
	      if (loop with state = (disambiguate-setup
				     operator #'ALWAYS-AFTER)
		       do (setf state (disambiguate-forward operator state))
			  (graphics-add-step *plan* level)
		          (multiple-value-setq
			   (new-adds new-dels)
			   (ua-adjust-unachieved-goals operator complete-cl))
			  (if *debug* (ua-order-plan))
			  (incf *visits*)
			  (if (ua-extend-fast complete-cl (1+ level))
			      (return t))
			  (setf state (disambiguate-backtrack operator state))
			  (readjust-unach-goals complete-cl new-adds new-dels)
			  (if *back-halt* (break))
			  (graphics-del-step *plan* level)
		       until (null state))
	      do (return t)
	      else do (progn (remove-ordering operator (pred-op pre))
			     (remove-ordering deleter operator)
			     (destroy-operator-instance)))))))


(defun mod1-ua-extend-fast (complete-cl &optional (level 0))
;;  (if (and (eq level 5) (zerop (new-random 10))) (break))
  (cond
   ((null (caar complete-cl)) t)
   ((not (< level *depth-level*)) nil)
   ((> *visits* *node-limit*) nil)
   (t (let ((pre (select-goal (first complete-cl) level))
	    (new-adds nil) (new-dels nil))
	(if *debug* (format *debug* "G: ~A~%" (pred-goal pre)))
	(loop with operator = nil
	      with deleter  = nil
	      with operators = (achievement-operators (pred-goal pre))
	      initially (if (eql *random-goal* 'random)
			    (shuffle-list operators))
	      for operator-class in operators
	      do (setf operator (create-operator-instance operator-class))
	         (setf deleter (find-deleter pre))
		 (add-ordering deleter operator)
		 (add-ordering operator (pred-op pre))
		 (if *debug* (format *debug* "OP: ~A~%"
			 (operator-class-name (operator-class operator))))
	      if (loop with state = (disambiguate-setup operator #'determine-order)
		       do (setf state (disambiguate-forward operator state))
			  (graphics-add-step *plan* level)
		          (multiple-value-setq
			   (new-adds new-dels)
			   (ua-adjust-unachieved-goals operator complete-cl))
			  (if *debug* (ua-order-plan))
			  (incf *visits*)
			  (cond ((mod1-ua-extend-fast complete-cl (1+ level))
				 (return t))
				(t (setf state (disambiguate-backtrack operator state))
				   (readjust-unach-goals
				    complete-cl new-adds new-dels)))
			  (graphics-del-step *plan* level)
		       until (null state))
	      do (return t)
	      else do (progn (remove-ordering operator (pred-op pre))
			     (remove-ordering deleter operator)
			     (destroy-operator-instance)))))))


(defun copy-stack-from-state (state)
  (let ((stack (car state)))
    (loop for (step (dir . rest1) . rest2) in stack
	  collect (list step dir))))


(defun reinstate (op stack)
  (if stack (reinstate op (cdr stack)))
  (case (nth 1 (car stack))
	(after (add-ordering (caar stack) op))
	(before (add-ordering op (caar stack)))))


(defun deinstate (op stack)
  (if stack
      (progn (case (nth 1 (car stack))
		   (after (remove-ordering (caar stack) op))
		   (before (remove-ordering op (caar stack))))
	     (deinstate op (cdr stack)))))
    

(defun mod2-ua-extend-fast (complete-cl &optional (level 0))
  (cond
   ((null (caar complete-cl)) t)
   ((not (< level *depth-level*)) nil)
   ((> *visits* *node-limit*) nil)
   (t (let ((pre (select-goal (first complete-cl) level)) (op nil)
	    (deleter nil) (operators nil) (heuristic nil))
	(if *debug* (format *debug* "G: ~A~%" (pred-goal pre)))
	(setf operators (achievement-operators (pred-goal pre)))
	(if (eql *random-goal* 'random) (shuffle-list operators))
	(dolist (op-class operators)
	  (setf op (create-operator-instance op-class)
		deleter (find-deleter pre)
		heuristic nil)
	  (add-ordering deleter op)
	  (add-ordering op (pred-op pre))
	  (if *debug* (format *debug* "OP: ~A~%"
			      (operator-class-name (operator-class op))))
	  (loop with state = (disambiguate-setup op #'always-after)
		with flag  = (if (nth 1 state) t)
		with stack = nil
		while (setf state (disambiguate-forward op state)) do
		(multiple-value-bind
		 (n-adds n-dels)
		 (ua-adjust-unachieved-goals op complete-cl)
		 (let ((h (- (length n-dels) (length n-adds))))
		   (loop for pre in (operator-pre op)
			 if (not (pred-? pre)) do (incf h))
		   (setf stack (if flag (copy-stack-from-state state) t))
		   (setf heuristic (insert-ordering h stack heuristic)))
		 (if flag (setf state (disambiguate-backtrack op state)))
		 (readjust-unach-goals complete-cl n-adds n-dels))
		while (and state flag))
	  (setf heuristic (sort heuristic #'< :key #'car))
	  (if (loop
	       with stack = nil
	       while (setf stack (extract-ordering heuristic)) do
	       (if (eql stack t) (setf stack nil))
	       (reinstate op stack)
	       (graphics-add-step *plan* level)
	       (multiple-value-bind
		(n-adds n-dels) (ua-adjust-unachieved-goals op complete-cl)
		(if *debug* (progn
			      (format t "U: ~D, A: ~D, " (length n-dels)
				      (length n-adds)) (ua-order-plan)))
		(incf *visits*)
		(if (mod2-ua-extend-fast complete-cl (1+ level)) (return t))
		(deinstate op stack)
		(readjust-unach-goals complete-cl n-adds n-dels)
		(graphics-del-step *plan* level))
	       while stack)
	      (return t))
	  (remove-ordering op (pred-op pre))
	  (remove-ordering deleter op)
	  (destroy-operator-instance))))))
    

(defun mod3-ua-extend-fast (complete-cl &optional (level 0))
  (cond
   ((null (caar complete-cl)) t)
   ((not (< level *depth-level*)) nil)
   ((> *visits* *node-limit*) nil)
   (t (let ((pre (select-goal (first complete-cl) level)) (op nil)
	    (deleter nil) (heuristic nil))
	(if *debug* (format *debug* "G: ~A~%" (pred-goal pre)))
	(dolist (op-class (achievement-operators (pred-goal pre)))
	  (setf op (create-operator-instance op-class)
		deleter (find-deleter pre))
	  (add-ordering deleter op)
	  (add-ordering op (pred-op pre))
	  (loop with state = (disambiguate-setup op #'always-after)
		with flag  = (if (nth 1 state) t)
		with stack = nil
		with stuff = nil
		while (setf state (disambiguate-forward op state)) do
		(multiple-value-bind
		 (n-adds n-dels)
		 (ua-adjust-unachieved-goals op complete-cl)
		 (let ((h (- (length n-dels) (length n-adds))))
		   (loop for pre in (operator-pre op)
			 if (not (pred-? pre)) do (incf h))
		   (setf stack (if flag (copy-stack-from-state state) nil)
			 stuff (list op-class stack deleter))
		   (setf heuristic (insert-ordering h stuff heuristic)))
		 (if flag (setf state (disambiguate-backtrack op state)))
		 (readjust-unach-goals complete-cl n-adds n-dels))
		while (and state flag))
	  (remove-ordering op (pred-op pre))
	  (remove-ordering deleter op)
	  (destroy-operator-instance))
	(setf heuristic (sort heuristic #'< :key #'car))
	(loop with tuple = nil
	      with stack = nil
	      while (setf tuple (extract-ordering heuristic)) do
	      (setf op (create-operator-instance (nth 0 tuple))
		    stack (nth 1 tuple)
		    deleter (nth 2 tuple))
	      (add-ordering deleter op)
	      (add-ordering op (pred-op pre))
	      (if *debug* (format *debug* "OP: ~A~%"
				  (operator-class-name (operator-class op))))
	      (reinstate op stack)
	      (graphics-add-step *plan* level)
	      (multiple-value-bind
	       (n-adds n-dels) (ua-adjust-unachieved-goals op complete-cl)
	       (if *debug* (progn (format t "U: ~D, A: ~D, " (length n-dels)
					  (length n-adds)) (ua-order-plan)))
	       (incf *visits*)
	       (if (mod3-ua-extend-fast complete-cl (1+ level)) (return t))
	       (deinstate op stack)
	       (readjust-unach-goals complete-cl n-adds n-dels)
	       (graphics-del-step *plan* level))
	      (remove-ordering op (pred-op pre))
	      (remove-ordering deleter op)
	      (destroy-operator-instance))))))


(defun uas nil
  (ua)
  (let ((result (ua-extend-slow (ua-get-unachieved-goals))))
    (format *out-stream* "Number of calls: ~A~%" *visits*)
    (dolist (i *plan*) (format *out-stream* "~A~%" i))
    result))


(defun uaf nil
  (ua)
  (let ((t1 (get-internal-run-time))
	(result (ua-extend-fast (ua-get-unachieved-goals)))
	(t2 (get-internal-run-time)))
    (format *out-stream* "Number of calls: ~A~%" *visits*)
    (dolist (i *plan*) (format *out-stream* "~A~%" i))
    (pprint *num-sols*)
    (pprint *num-fails*)
    `(,result ,*visits* ,(/ (- t2 t1) internal-time-units-per-second 1.0)
	      ,*num-sols* ,*num-fails* )))


;; Uses point to point (or MOD1) type heuristic
(defun puaf nil
  (ua)
  (let ((t1 (get-internal-run-time))
	(result (mod1-ua-extend-fast (ua-get-unachieved-goals)))
	(t2 (get-internal-run-time)))
    (format *out-stream* "Number of calls: ~A~%" *visits*)
    (dolist (i *plan*) (format *out-stream* "~A~%" i))
    `(,result ,*visits* ,(/ (- t2 t1) internal-time-units-per-second 1.0))))


;; Uses complete (or MOD2) type heuristic (looks at all possibilities)
(defun cuaf nil
  (ua)
  (let ((t1 (get-internal-run-time))
	(result (mod2-ua-extend-fast (ua-get-unachieved-goals)))
	(t2 (get-internal-run-time)))
    (format *out-stream* "Number of calls: ~A~%" *visits*)
    (dolist (i *plan*) (format *out-stream* "~A~%" i))
    `(,result ,*visits* ,(/ (- t2 t1) internal-time-units-per-second 1.0))))


;; Uses complete (or MOD3) type heuristic (looks at all possibilities)
(defun fuaf nil
  (ua)
  (let ((t1 (get-internal-run-time))
	(result (mod3-ua-extend-fast (ua-get-unachieved-goals)))
	(t2 (get-internal-run-time)))
    (format *out-stream* "Number of calls: ~A~%" *visits*)
    (dolist (i *plan*) (format *out-stream* "~A~%" i))
    `(,result ,*visits* ,(/ (- t2 t1) internal-time-units-per-second 1.0))))


(defun buaf nil
  (let* ((original *depth-level*)
	 (result (loop for depth from 1 upto original
		       do (setf *depth-level* depth)
		       (ua)
		       if (ua-extend-fast (ua-get-unachieved-goals))
		       do (return t))))
    (format *out-stream* "Number of calls: ~A~%" *visits*)
    (if result (dolist (i *plan* t) (format *out-stream* "~A~%" i)))))


(defun ua-order-plan nil
  (clear-out-all-marks 0)  
  (loop with collect = nil
	with linearization = nil
	with linear? = t
	with continue = t
	while continue
	do (setf collect nil)
	   (dolist (op *plan*)
		   (if (= (length (operator-preds op)) (operator-mark op))
		       (push op collect)))
	   (case (length collect)
		 (0 (if (not (= (length *plan*) (length linearization)))
			(error "Whoops"))
		    (setf continue nil))
		 (1)
		 (t (setf linear? nil)))
	   (cond (continue
		  (push (car collect) linearization)
		  (print (car collect))
		  (incf (operator-mark (car collect)))
		  (dolist (temp (operator-succs (car collect)))
			  (incf (operator-mark temp)))))
	finally (format t "~a~%" (nreverse linearization))
;		(format t "~:[Not ~;~]Totally Ordered~%" linear?)
	))


