;;;  Contains: Functions which run TO with iterative sample approach
;;;  Author:  Andy Philips (abp@goedel.arc.nasa.gov)
;;;  Created: Thu Dec  5 15:42:06 1991

#|

This file defines functions that allow TO to perform Monte Carlo
simulations (shotgun) on the search tree for a problem.  It requires
all of the functions that TO normally requires to solve problems and
the same problem specifications.

[This strategy is now referred to as "iterative sampling" in our papers"]

To use the functions defined within call:

(to-shotgun <iterations>)

Notes:

(NEW-RANDOM x) is the same as (RANDOM X)

This system requires the LOOP macro.

Try loading problems from the P/ directory.  Try 'problems.lisp' and
call the function '(PROBLEM 1)'.

These files work best when compiled.

(setf *graphics-on* nil) when not in ALLEGRO with common windows.

|#

(require :loop)

(proclaim '(optimize (speed 3) (safety 0)))

(proclaim '(special *initial-operator* *final-operator* *depth-level*
		    *visits* *out-stream* *debug* *back-halt* *node-limit*))


(defun shotgun-to-extend-fast (complete-cl &optional (level 0))
  (cond
   ((null (caar complete-cl)) t)
   ((not (< level *depth-level*)) nil)
   (t (let ((goal-to-achieve (select-goal (first complete-cl) level))
	    (new-adds nil) (new-dels nil) (operator nil) (heuristic nil)
	    (position nil))
	(if *debug* (format *debug* "G: ~A~%" (pred-goal goal-to-achieve)))
	(setf operator (create-operator-instance
			(choose-item-randomly (achievement-operators
					       (pred-goal goal-to-achieve))))
	      heuristic nil)
	(if *debug* (format *debug* "OP: ~A~%"
			    (operator-class-name (operator-class operator))))
	(loop with position = (pred-op goal-to-achieve)
	      do (setf position (car (operator-preds position)))
	      while position
	      do    (push position heuristic)
	      until (member (pred-goal goal-to-achieve)
			    (operator-class-del
			     (operator-class position))
			    :test #'equal))
	(if (not (setf position (choose-item-randomly heuristic)))
	    (error "No place to put operator!"))
	(place-operator-after operator position)
	(graphics-add-step *plan* level)
	(multiple-value-setq
	 (new-adds new-dels) (adjust-unachieved-goals operator complete-cl))
	(if *debug* (format *debug* "~A ~A~%" *plan* position))
	(incf *visits*)
	(shotgun-to-extend-fast complete-cl (1+ level))))))


(defun shotgun1-to-extend-fast (complete-cl &optional (level 0))
  (cond
   ((null (caar complete-cl)) t)
   ((not (< level *depth-level*)) nil)
   (t (let ((goal-to-achieve (select-goal (first complete-cl) level)) (op nil)
	    (operators nil) (heuristic nil) (tuple nil) (position nil))
	(if *debug* (format *debug* "G: ~A~%" (pred-goal goal-to-achieve)))
	(setf operators (achievement-operators (pred-goal goal-to-achieve)))
	(if (eql *random-goal* 'random) (shuffle-list operators))
	(dolist (op-class operators nil)
	  (setf op (create-operator-instance op-class))
	  (loop with position = (pred-op goal-to-achieve)
		while (setf position (car (operator-preds position))) do
		(place-operator-after op position)
		(multiple-value-bind
		 (n-adds n-dels) (adjust-unachieved-goals op complete-cl)
		 (let ((h (- (length n-dels) (length n-adds))))
		   (dolist (pre (operator-pre op))
			   (if (not (pred-? pre)) (incf h)))
		   (setf heuristic
			 (insert-ordering h `(,op-class ,position) heuristic)))
		 (remove-operator-from-plan op)
		 (readjust-unach-goals complete-cl n-adds n-dels))
		until (member (pred-goal goal-to-achieve)
			      (operator-class-del (operator-class position))
			      :test #'equal))
	  (destroy-operator-instance))
	(setf heuristic (sort heuristic #'< :key #'car))
	(setf tuple (extract-ordering heuristic))
	(if tuple
	    (progn
	      (setf op (create-operator-instance (car tuple))
		    position (nth 1 tuple))
	      (if *debug* (format *debug* "OP: ~A~%"
				  (operator-class-name (operator-class op))))
	      (place-operator-after op position)
	      (graphics-add-step *plan* level)
	      (multiple-value-bind
	       (n-adds n-dels) (adjust-unachieved-goals op complete-cl)
	       (declare (ignore n-dels n-adds))
	       (if *debug* (format *debug* "~A ~A~%" *plan* position))
	       (incf *visits*)
	       (shotgun1-to-extend-fast complete-cl (1+ level)))))))))


(defun shotgun2-to-extend-fast (complete-cl &optional (level 0))
  (cond
   ((null (caar complete-cl)) t)
   ((not (< level *depth-level*)) nil)
   (t (let ((goal-to-achieve (select-goal (first complete-cl) level)) (op nil)
	    (operators nil) (heuristic nil) (tuple nil) (position nil))
	(if *debug* (format *debug* "G: ~A~%" (pred-goal goal-to-achieve)))
	(setf operators (achievement-operators (pred-goal goal-to-achieve)))
	(if (eql *random-goal* 'random) (shuffle-list operators))
	(dolist (op-class operators nil)
	  (setf op (create-operator-instance op-class))
	  (loop with position = (pred-op goal-to-achieve)
		while (setf position (car (operator-preds position))) do
		(push `(,op-class ,position) heuristic)
		until (member (pred-goal goal-to-achieve)
			      (operator-class-del (operator-class position))
			      :test #'equal))
	  (destroy-operator-instance))
	(if (setf tuple (choose-item-randomly heuristic))
	    (progn
	      (setf op (create-operator-instance (car tuple))
		    position (nth 1 tuple))
	      (if *debug* (format *debug* "OP: ~A~%"
				  (operator-class-name (operator-class op))))
	      (place-operator-after op position)
	      (graphics-add-step *plan* level)
	      (multiple-value-bind
	       (n-adds n-dels) (adjust-unachieved-goals op complete-cl)
	       (declare (ignore n-dels n-adds))
	       (if *debug* (format *debug* "~A ~A~%" *plan* position))
	       (incf *visits*)
	       (shotgun2-to-extend-fast complete-cl (1+ level)))))))))


(defun calc-to-shotgun-extend-fast (complete-cl &optional (level 0))
  (cond
   ((null (caar complete-cl)) 1.0)
   ((not (< level *depth-level*)) 0.0)
   (t (let ((op nil)
	    (operators nil) (heuristic nil) (success 0.0) (dlh 0) (dag 0)
	    (all-goals (complete-list-of-goals (first complete-cl))))
	(setf dag (length all-goals))
	(loop
	 for goal-to-achieve in all-goals do
	 (if *debug* (format *debug* "G: ~A~%" (pred-goal goal-to-achieve)))
	 (setf operators (achievement-operators (pred-goal goal-to-achieve))
	       heuristic nil)
	 (dolist (op-class operators nil)
	   (setf op (create-operator-instance op-class))
	   (loop with position = (pred-op goal-to-achieve)
		 while (setf position (car (operator-preds position))) do
		 (place-operator-after op position)
		 (multiple-value-bind
		  (n-adds n-dels) (adjust-unachieved-goals op complete-cl)
		  (let ((h (- (length n-dels) (length n-adds))))
		    (dolist (pre (operator-pre op))
			    (if (not (pred-? pre)) (incf h)))
		    (setf heuristic
			  (insert-ordering h `(,op-class ,position) heuristic)))
		  (remove-operator-from-plan op)
		  (readjust-unach-goals complete-cl n-adds n-dels))
		 until (member (pred-goal goal-to-achieve)
			       (operator-class-del (operator-class position))
			       :test #'equal))
	   (destroy-operator-instance))
	 (setf heuristic (sort heuristic #'< :key #'car))
	 (rplacd heuristic nil)
	 (setf dlh (length (cdar heuristic)))
	 (loop with tuple = nil
	       with position = nil
	       while (setf tuple (extract-ordering heuristic)) do
	       (setf op (create-operator-instance (car tuple))
		     position (nth 1 tuple))
	       (if *debug* (format *debug* "OP: ~A~%"
				   (operator-class-name (operator-class op))))
	       (place-operator-after op position)
	       (graphics-add-step *plan* level)
	       (multiple-value-bind
		(n-adds n-dels) (adjust-unachieved-goals op complete-cl)
		(if *debug* (format *debug* "~A ~A~%" *plan* position))
		(incf *visits*)
		(incf success (/ (calc-to-shotgun-extend-fast
				  complete-cl (1+ level)) dlh))
		(remove-operator-from-plan op)
		(readjust-unach-goals complete-cl n-adds n-dels)
		(graphics-del-step *plan* level))
	       (destroy-operator-instance)))
	(/ success dag)))))


(defun analyze-shotgun-result (l)
  (loop with len = (length l)
	with success = 0
	with time = 0
	with visits = 0
	for (s v z) in l
	do (if s (incf success))
	   (incf time z)
	   (incf visits v)
	finally
	(format t "Number of trials: ~d~%~
                   Success Rate:     ~f~%~
		   Average Time:     ~f~%~
		   Average Depth:    ~F~%"
		len (/ success len 1.0) (/ time len 1.0) (/ visits len 1.0)))
  l)


(defun to-shotgun (attempts)
  (let ((str *out-stream*) (gs *random-goal*) (t1 0) (t2 0) (result nil))
    (setf *out-stream* nil *random-goal* 'random)
    (setf result
	  (loop
	   for count from 1 to attempts
	   do (to)
	      (setf t1 (get-internal-run-time))
	      (setf result (shotgun-to-extend-fast (get-unachieved-goals)))
	      (setf t2 (get-internal-run-time))
	   collect `(,result ,*visits*
		     ,(/ (- t2 t1) internal-time-units-per-second 1.0))))
    (setf *out-stream* str *random-goal* gs)
    (analyze-shotgun-result result)))


(defun to-shotgun-alt (attempts)
  (let ((str *out-stream*) (gs *random-goal*) (t1 0) (t2 0) (result nil))
    (setf *out-stream* nil *random-goal* 'random)
    (setf result
	  (loop
	   for count from 1 to attempts
	   do (to)
	      (setf t1 (get-internal-run-time))
	      (setf result (shotgun2-to-extend-fast (get-unachieved-goals)))
	      (setf t2 (get-internal-run-time))
	   collect `(,result ,*visits*
		     ,(/ (- t2 t1) internal-time-units-per-second 1.0))))
    (setf *out-stream* str *random-goal* gs)
    (analyze-shotgun-result result)))


(defun to-shotgun-mg (attempts)
  (let ((str *out-stream*) (gs *random-goal*) (t1 0) (t2 0) (result nil))
    (setf *out-stream* nil *random-goal* 'random)
    (setf result
	  (loop
	   for count from 1 to attempts
	   do (to)
	      (setf t1 (get-internal-run-time))
	      (setf result (shotgun1-to-extend-fast (get-unachieved-goals)))
	      (setf t2 (get-internal-run-time))
	   collect `(,result ,*visits*
		     ,(/ (- t2 t1) internal-time-units-per-second 1.0))))
    (setf *out-stream* str *random-goal* gs)
    (analyze-shotgun-result result)))


(defun to-shotgun-prime (&optional (max-attempts nil))
  (if max-attempts
      (format t "This is no longer used.  It is now calculated~%"))
  (setf max-attempts (ceiling *node-limit* *depth-level*))
  (let ((str *out-stream*) (gs *random-goal*) (t1 0) (t2 0) (result nil))
    (setf *out-stream* nil *random-goal* 'random)
    (setf result
	  (loop
	   with tt = 0
	   with nodes = 0
	   for count from 1 to max-attempts
	   do (to)
	      (setf t1 (get-internal-run-time))
	      (setf result (shotgun-to-extend-fast (get-unachieved-goals)))
	      (setf t2 (get-internal-run-time))
	      (incf nodes *visits*)
	      (incf tt (/ (- t2 t1) internal-time-units-per-second 1.0))
	   until result
	   finally (return (list result nodes tt))))
    (setf *out-stream* str *random-goal* gs)
    result))


(defun to-shotgun-alt-prime (&optional (max-attempts nil))
  (if max-attempts
      (format t "This is no longer used.  It is now calculated~%"))
  (setf max-attempts (ceiling *node-limit* *depth-level*))
  (let ((str *out-stream*) (gs *random-goal*) (t1 0) (t2 0) (result nil))
    (setf *out-stream* nil *random-goal* 'random)
    (setf result
	  (loop
	   with tt = 0
	   with nodes = 0
	   for count from 1 to max-attempts
	   do (to)
	      (setf t1 (get-internal-run-time))
	      (setf result (shotgun2-to-extend-fast (get-unachieved-goals)))
	      (setf t2 (get-internal-run-time))
	      (incf nodes *visits*)
	      (incf tt (/ (- t2 t1) internal-time-units-per-second 1.0))
	   until result
	   finally (return (list result nodes tt))))
    (setf *out-stream* str *random-goal* gs)
    result))


(defun to-shotgun-mg-prime (&optional (max-attempts nil))
  (if max-attempts
      (format t "This is no longer used.  It is now calculated~%"))
  (setf max-attempts (ceiling *node-limit* *depth-level*))
  (let ((str *out-stream*) (gs *random-goal*) (t1 0) (t2 0) (result nil))
    (setf *out-stream* nil *random-goal* 'random)
    (setf result
	  (loop
	   with tt = 0
	   with nodes = 0
	   for count from 1 to max-attempts
	   do (to)
	      (setf t1 (get-internal-run-time))
	      (setf result (shotgun1-to-extend-fast (get-unachieved-goals)))
	      (setf t2 (get-internal-run-time))
	      (incf nodes *visits*)
	      (incf tt (/ (- t2 t1) internal-time-units-per-second 1.0))
	   until result
	   finally (return (list result nodes tt))))
    (setf *out-stream* str *random-goal* gs)
    result))


(defun to-shotgun-complete-search nil
  (to)
  (let ((t1 (get-internal-run-time))
	(result (calc-to-shotgun-extend-fast (get-unachieved-goals)))
	(t2 (get-internal-run-time)))
    (format *out-stream* "Number of calls: ~A~%" *visits*)
    (dolist (i *plan*) (format *out-stream* "~A~%" i))
    `(,result ,*visits* ,(/ (- t2 t1) internal-time-units-per-second 1.0))))
