;;;  Contains: Functions which run UA with iterative sample approach
;;;  Author:  Andy Philips (abp@goedel.arc.nasa.gov)
;;;  Created: Fri Dec  6 15:32:06 1991


#|
See comments in shotgun-TO.lisp.
[This strategy is now referred to as "iterative sampling" in our papers"]


Commands to call from within this system are:

(ua-shotgun1 <iterations>)	Follows line of disambiguation
(ua-shotgun2 <iterations>)	Randomly chooses operator position
(ua-shotgun-alt <iterations>)	Randomly chooses from position X operator
(ua-shotgun-mg <iterations>)	MinGoals on shotgun
(ua-shotgun-mg-prime)		Expected Nodes with MinGoals
(ua-shotgun-prime)		Expected Nodes possibly skewed distribution
(ua-shotgun-alt-prime)		Expected Nodes possibly unskewed distribution


Notes:

(NEW-RANDOM x) is the same as (RANDOM X)

This system requires the LOOP macro.

Try loading problems from the P/ directory.  Try 'problems.lisp' and
call the function '(PROBLEM 1)'.

These files work best when compiled.

(setf *graphics-on* nil) when not in ALLEGRO with common windows.

|#

;;(proclaim '(optimize (speed 3) (safety 0)))


(proclaim '(special *initial-operator* *final-operator* *depth-level*
		    *visits* *out-stream* *debug* *back-halt* *node-limit*))


(defun shotgun1-ua-extend-fast (complete-cl &optional (level 0))
  (cond
   ((null (caar complete-cl)) t)
   ((not (< level *depth-level*)) nil)
   (t (let ((pre (select-goal (first complete-cl) level)) (state nil)
	    (new-adds nil) (new-dels nil) (operator nil) (deleter nil))
	(if *debug* (format *debug* "G: ~A~%" (pred-goal pre)))
	(setf operator (create-operator-instance
			(choose-item-randomly
			 (achievement-operators (pred-goal pre)))))
	(setf deleter (find-deleter pre))
	(add-ordering deleter operator)
	(add-ordering operator (pred-op pre))
	(if *debug* (format *debug* "OP: ~A~%"
			    (operator-class-name (operator-class operator))))
	(setf state (disambiguate-setup operator #'random-orderings))
	(setf (nth 1 state) (shuffle-list (nth 1 state)))
	(setf state (disambiguate-forward operator state))
	(graphics-add-step *plan* level)
	(multiple-value-setq
	 (new-adds new-dels)
	 (ua-adjust-unachieved-goals operator complete-cl))
	(if *debug* (ua-order-plan))
	(incf *visits*)
	(shotgun1-ua-extend-fast complete-cl (1+ level))))))
    

(defun random-orderings (&rest y)
  (declare (ignore y))
  (if (zerop (new-random 2)) 'before 'after))


(defun shotgun2-ua-extend-fast (complete-cl &optional (level 0))
  (cond
   ((null (caar complete-cl)) t)
   ((not (< level *depth-level*)) nil)
   (t (let ((pre (select-goal (first complete-cl) level))
	    (heuristic nil) (operator nil) (deleter nil))
	(if *debug* (format *debug* "G: ~A~%" (pred-goal pre)))
	(setf operator (create-operator-instance
			(choose-item-randomly
			 (achievement-operators (pred-goal pre))))
	      deleter (find-deleter pre))
	(add-ordering deleter operator)
	(add-ordering operator (pred-op pre))
	(if *debug* (format *debug* "OP: ~A~%"
			    (operator-class-name (operator-class operator))))
	(loop with state = (disambiguate-setup operator #'always-after)
	      initially (if (not (nth 1 state)) (return))
	      while (setf state (disambiguate-forward operator state))
	      do (multiple-value-bind
		  (new-adds new-dels)
		  (ua-adjust-unachieved-goals operator complete-cl)
		  (if *debug* (ua-order-plan))
		  (push (copy-stack-from-state state) heuristic)
		  (setf state (disambiguate-backtrack operator state))
		  (readjust-unach-goals complete-cl new-adds new-dels))
	      while state)
	(reinstate operator (choose-item-randomly heuristic))
	(graphics-add-step *plan* level)
	(multiple-value-bind
	 (new-adds new-dels)
	 (ua-adjust-unachieved-goals operator complete-cl)
	 (if *debug* (progn (format t "U: ~D, A: ~D, " (length new-dels)
				    (length new-adds))
			    (ua-order-plan)))
	 (incf *visits*)
	 (shotgun2-ua-extend-fast complete-cl (1+ level)))))))


(defun shotgun3-ua-extend-fast (complete-cl &optional (level 0))
  (cond
   ((null (caar complete-cl)) t)
   ((not (< level *depth-level*)) nil)
   (t (let ((pre (select-goal (first complete-cl) level)) (op nil)
	    (deleter nil) (heuristic nil))
	(if *debug* (format *debug* "G: ~A~%" (pred-goal pre)))
	(dolist (op-class (achievement-operators (pred-goal pre)))
	  (setf op (create-operator-instance op-class)
		deleter (find-deleter pre))
	  (add-ordering deleter op)
	  (add-ordering op (pred-op pre))
	  (loop with state = (disambiguate-setup op #'always-after)
		with flag  = (if (nth 1 state) t)
		with stack = nil
		with stuff = nil
		while (setf state (disambiguate-forward op state)) do
		(multiple-value-bind
		 (n-adds n-dels)
		 (ua-adjust-unachieved-goals op complete-cl)
		 (let ((h (- (length n-dels) (length n-adds))))
		   (loop for pre in (operator-pre op)
			 if (not (pred-? pre)) do (incf h))
		   (setf stack (if flag (copy-stack-from-state state) nil)
			 stuff (list op-class stack deleter))
		   (setf heuristic (insert-ordering h stuff heuristic)))
		 (if flag (setf state (disambiguate-backtrack op state)))
		 (readjust-unach-goals complete-cl n-adds n-dels))
		while (and state flag))
	  (remove-ordering op (pred-op pre))
	  (remove-ordering deleter op)
	  (destroy-operator-instance))
	(setf heuristic (sort heuristic #'< :key #'car))
	(let ((tuple nil) (stack nil))
	  (setf tuple (extract-ordering heuristic))
	  (if tuple
	      (progn
		(setf op (create-operator-instance (nth 0 tuple))
		      stack (nth 1 tuple)
		      deleter (nth 2 tuple))
		(add-ordering deleter op)
		(add-ordering op (pred-op pre))
		(if *debug* (format *debug* "OP: ~A~%"
				    (operator-class-name (operator-class op))))
		(reinstate op stack)
		(graphics-add-step *plan* level)
		(multiple-value-bind
		 (n-adds n-dels) (ua-adjust-unachieved-goals op complete-cl)
		 (if *debug* (progn (format t "U: ~D, A: ~D, " (length n-dels)
					    (length n-adds)) (ua-order-plan)))
		 (incf *visits*)
		 (shotgun3-ua-extend-fast complete-cl (1+ level))))))))))


(defun shotgun4-ua-extend-fast (complete-cl &optional (level 0))
  (cond
   ((null (caar complete-cl)) t)
   ((not (< level *depth-level*)) nil)
   (t (let ((pre (select-goal (first complete-cl) level)) (op nil)
	    (deleter nil) (heuristic nil))
	(if *debug* (format *debug* "G: ~A~%" (pred-goal pre)))
	(dolist (op-class (achievement-operators (pred-goal pre)))
	  (setf op (create-operator-instance op-class)
		deleter (find-deleter pre))
	  (add-ordering deleter op)
	  (add-ordering op (pred-op pre))
	  (loop with state = (disambiguate-setup op #'always-after)
		with flag  = (if (nth 1 state) t)
		with stack = nil
		with stuff = nil
		while (setf state (disambiguate-forward op state)) do
		(setf stack (if flag (copy-stack-from-state state) nil)
		      stuff (list op-class stack deleter))
		(push stuff heuristic)
		(if flag (setf state (disambiguate-backtrack op state)))
		while (and state flag))
	  (remove-ordering op (pred-op pre))
	  (remove-ordering deleter op)
	  (destroy-operator-instance))
	(let ((tuple nil) (stack nil))
	  (setf tuple (choose-item-randomly heuristic))
	  (if tuple
	      (progn
		(setf op (create-operator-instance (nth 0 tuple))
		      stack (nth 1 tuple)
		      deleter (nth 2 tuple))
		(add-ordering deleter op)
		(add-ordering op (pred-op pre))
		(if *debug* (format *debug* "OP: ~A~%"
				    (operator-class-name (operator-class op))))
		(reinstate op stack)
		(graphics-add-step *plan* level)
		(multiple-value-bind
		 (n-adds n-dels) (ua-adjust-unachieved-goals op complete-cl)
		 (if *debug* (progn (format t "U: ~D, A: ~D, " (length n-dels)
					    (length n-adds)) (ua-order-plan)))
		 (incf *visits*)
		 (shotgun4-ua-extend-fast complete-cl (1+ level))))))))))


(defun calc-ua-shotgun-extend-fast (complete-cl &optional (level 0))
  (cond
   ((null (caar complete-cl)) 1.0)
   ((not (< level *depth-level*)) 0.0)
   (t (let ((op nil) (deleter nil) (heuristic nil) (success 0.0) (dlh 0)
	    (all-goals (complete-list-of-goals (first complete-cl))))
	(loop
	 for pre in all-goals do
	 (if *debug* (format *debug* "G: ~A~%" (pred-goal pre)))
	 (setf heuristic nil)
	 (dolist (op-class (achievement-operators (pred-goal pre)))
	   (setf op (create-operator-instance op-class)
		 deleter (find-deleter pre))
	   (add-ordering deleter op)
	   (add-ordering op (pred-op pre))
	   (loop with state = (disambiguate-setup op #'always-after)
		 with flag  = (if (nth 1 state) t)
		 with stack = nil
		 with stuff = nil
		 while (setf state (disambiguate-forward op state)) do
		 (multiple-value-bind
		  (n-adds n-dels)
		  (ua-adjust-unachieved-goals op complete-cl)
		  (let ((h (- (length n-dels) (length n-adds))))
		    (loop for pre in (operator-pre op)
			  if (not (pred-? pre)) do (incf h))
		    (setf stack (if flag (copy-stack-from-state state) nil)
			  stuff (list op-class stack deleter))
		    (setf heuristic (insert-ordering h stuff heuristic)))
		  (if flag (setf state (disambiguate-backtrack op state)))
		  (readjust-unach-goals complete-cl n-adds n-dels))
		 while (and state flag))
	   (remove-ordering op (pred-op pre))
	   (remove-ordering deleter op)
	   (destroy-operator-instance))
	 (setf heuristic (sort heuristic #'< :key #'car))
	 (rplacd heuristic nil)
	 (setf dlh (length (cdar heuristic)))
	 (loop with tuple = nil
	       with stack = nil
	       while (setf tuple (extract-ordering heuristic)) do
	       (setf op (create-operator-instance (nth 0 tuple))
		     stack (nth 1 tuple)
		     deleter (nth 2 tuple))
	       (add-ordering deleter op)
	       (add-ordering op (pred-op pre))
	       (if *debug* (format *debug* "OP: ~A~%"
				   (operator-class-name (operator-class op))))
	       (reinstate op stack)
	       (graphics-add-step *plan* level)
	       (multiple-value-bind
		(n-adds n-dels) (ua-adjust-unachieved-goals op complete-cl)
		(if *debug* (progn (format t "U: ~D, A: ~D, " (length n-dels)
					   (length n-adds)) (ua-order-plan)))
		(incf *visits*)
		(incf success (/ (calc-ua-shotgun-extend-fast
				  complete-cl (1+ level)) dlh))
		(deinstate op stack)
		(readjust-unach-goals complete-cl n-adds n-dels)
		(graphics-del-step *plan* level))
	       (remove-ordering op (pred-op pre))
	       (remove-ordering deleter op)
	       (destroy-operator-instance)))
	(/ success (length all-goals))))))


(defun ua-shotgun1 (attempts)
  (let ((str *out-stream*) (gs *random-goal*) (t1 0) (t2 0) (result nil))
    (setf *out-stream* nil *random-goal* 'random)
    (setf result
	  (loop
	   for count from 1 to attempts
	   do (ua)
	      (setf t1 (get-internal-run-time))
	      (setf result (shotgun1-ua-extend-fast (get-unachieved-goals)))
	      (setf t2 (get-internal-run-time))
	   collect `(,result ,*visits*
		     ,(/ (- t2 t1) internal-time-units-per-second 1.0))))
    (setf *out-stream* str *random-goal* gs)
    (analyze-shotgun-result result)))


(defun ua-shotgun2 (attempts)
  (let ((str *out-stream*) (gs *random-goal*) (t1 0) (t2 0) (result nil))
    (setf *out-stream* nil *random-goal* 'random)
    (setf result
	  (loop
	   for count from 1 to attempts
	   do (ua)
	      (setf t1 (get-internal-run-time))
	      (setf result (shotgun2-ua-extend-fast (get-unachieved-goals)))
	      (setf t2 (get-internal-run-time))
	   collect `(,result ,*visits*
		     ,(/ (- t2 t1) internal-time-units-per-second 1.0))))
    (setf *out-stream* str *random-goal* gs)
    (analyze-shotgun-result result)))


(defun ua-shotgun-alt (attempts)
  (let ((str *out-stream*) (gs *random-goal*) (t1 0) (t2 0) (result nil))
    (setf *out-stream* nil *random-goal* 'random)
    (setf result
	  (loop
	   for count from 1 to attempts
	   do (ua)
	      (setf t1 (get-internal-run-time))
	      (setf result (shotgun4-ua-extend-fast (get-unachieved-goals)))
	      (setf t2 (get-internal-run-time))
	   collect `(,result ,*visits*
		     ,(/ (- t2 t1) internal-time-units-per-second 1.0))))
    (setf *out-stream* str *random-goal* gs)
    (analyze-shotgun-result result)))


(defun ua-shotgun-mg (attempts)
  (let ((str *out-stream*) (gs *random-goal*) (t1 0) (t2 0) (result nil))
    (setf *out-stream* nil *random-goal* 'random)
    (setf result
	  (loop
	   for count from 1 to attempts
	   do (ua)
	      (setf t1 (get-internal-run-time))
	      (setf result (shotgun3-ua-extend-fast (get-unachieved-goals)))
	      (setf t2 (get-internal-run-time))
	   collect `(,result ,*visits*
		     ,(/ (- t2 t1) internal-time-units-per-second 1.0))))
    (setf *out-stream* str *random-goal* gs)
    (analyze-shotgun-result result)))


(defun ua-shotgun-prime (&optional (max-attempts nil))
  (if max-attempts
      (format t "This is no longer used.  It is now calculated~%"))
  (setf max-attempts (ceiling *node-limit* *depth-level*))
  (let ((str *out-stream*) (gs *random-goal*) (t1 0) (t2 0) (result nil))
    (setf *out-stream* nil *random-goal* 'random)
    (setf result
	  (loop
	   with tt = 0
	   with nodes = 0
	   for count from 1 to max-attempts
	   do (ua)
	      (setf t1 (get-internal-run-time))
	      (setf result (shotgun1-ua-extend-fast (get-unachieved-goals)))
	      (setf t2 (get-internal-run-time))
	      (incf nodes *visits*)
	      (incf tt (/ (- t2 t1) internal-time-units-per-second 1.0))
	   until result
	   finally (return (list result nodes tt))))
    (setf *out-stream* str *random-goal* gs)
    result))


(defun ua-shotgun-alt-prime (&optional (max-attempts nil))
  (if max-attempts
      (format t "This is no longer used.  It is now calculated~%"))
  (setf max-attempts (ceiling *node-limit* *depth-level*))
  (let ((str *out-stream*) (gs *random-goal*) (t1 0) (t2 0) (result nil))
    (setf *out-stream* nil *random-goal* 'random)
    (setf result
	  (loop
	   with tt = 0
	   with nodes = 0
	   for count from 1 to max-attempts
	   do (ua)
	      (setf t1 (get-internal-run-time))
	      (setf result (shotgun4-ua-extend-fast (get-unachieved-goals)))
	      (setf t2 (get-internal-run-time))
	      (incf nodes *visits*)
	      (incf tt (/ (- t2 t1) internal-time-units-per-second 1.0))
	   until result
	   finally (return (list result nodes tt))))
    (setf *out-stream* str *random-goal* gs)
    result))


(defun ua-shotgun-mg-prime (&optional (max-attempts nil))
  (if max-attempts
      (format t "This is no longer used.  It is now calculated~%"))
  (setf max-attempts (ceiling *node-limit* *depth-level*))
  (let ((str *out-stream*) (gs *random-goal*) (t1 0) (t2 0) (result nil))
    (setf *out-stream* nil *random-goal* 'random)
    (setf result
	  (loop
	   with tt = 0
	   with nodes = 0
	   for count from 1 to max-attempts
	   do (ua)
	      (setf t1 (get-internal-run-time))
	      (setf result (shotgun3-ua-extend-fast (get-unachieved-goals)))
	      (setf t2 (get-internal-run-time))
	      (incf nodes *visits*)
	      (incf tt (/ (- t2 t1) internal-time-units-per-second 1.0))
	   until result
	   finally (return (list result nodes tt))))
    (setf *out-stream* str *random-goal* gs)
    result))


(defun ua-shotgun-complete-search nil
  (ua)
  (let ((t1 (get-internal-run-time))
	(result (calc-ua-shotgun-extend-fast (ua-get-unachieved-goals)))
	(t2 (get-internal-run-time)))
    (format *out-stream* "Number of calls: ~A~%" *visits*)
    (dolist (i *plan*) (format *out-stream* "~A~%" i))
    `(,result ,*visits* ,(/ (- t2 t1) internal-time-units-per-second 1.0))))